import struct
from typing import Optional

from bxutils.logging.log_level import LogLevel

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash

from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_BLOCK_HDR_SIZE, BTC_SHA_HASH_LEN
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import get_next_tx_size, pack_int_to_btc_varint, \
    pack_block_header


# FIXME, there's a lot of duplicate code between here and BlockHeader
class BlockBtcMessage(BtcMessage, AbstractBlockMessage):
    MESSAGE_TYPE = BtcMessageType.BLOCK

    def __init__(self, magic=None, version=None, prev_block=None, merkle_root=None,
                 timestamp=None, bits=None, nonce=None, txns=None, buf=None):
        if buf is None:
            total_tx_size = sum(len(tx) for tx in txns)
            buf = bytearray(BTC_HDR_COMMON_OFF + 80 + 9 + total_tx_size)
            self.buf = buf

            off = pack_block_header(buf, version, prev_block, merkle_root, timestamp, bits, nonce)
            off += pack_int_to_btc_varint(len(txns), buf, off)

            for tx in txns:
                buf[off:off + len(tx)] = tx
                off += len(tx)

            super(BlockBtcMessage, self).__init__(magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            assert not isinstance(buf, str)
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = None
        self._prev_block: Optional[BtcObjectHash] = None
        self._merkle_root = None
        self._bits = self._nonce = self._txn_count = self._txns = self._hash_val = None
        self._header = self._tx_offset = None
        self._timestamp = 0

    def log_level(self):
        return LogLevel.DEBUG

    def version(self):
        if self._version is None:
            off = BTC_HDR_COMMON_OFF
            self._version = struct.unpack_from('<I', self.buf, off)[0]
            off += 4
            self._prev_block = BtcObjectHash(self.buf, off, 32)
            off += 32
            self._merkle_root = BtcObjectHash(self.buf, off, 32)
            off += 32
            self._timestamp, self._bits, self._nonce = struct.unpack_from('<III', self.buf, off)
            off += 12
            self._txn_count, size = btc_varint_to_int(self.buf, off)
            off += size
            self._tx_offset = off

            self._header = self._memoryview[0:self._tx_offset]

        return self._version

    def prev_block_hash(self) -> BtcObjectHash:
        if self._version is None:
            self.version()

        assert self._prev_block is not None
        # pyre-fixme[7]: Expected `BtcObjectHash` but got `Optional[BtcObjectHash]`.
        return self._prev_block

    def merkle_root(self):
        if self._version is None:
            self.version()
        return self._merkle_root

    def timestamp(self) -> int:
        """
        :return: seconds since epoch
        """
        assert self._timestamp is not None
        if self._version is None:
            self.version()
        return self._timestamp

    def bits(self):
        if self._version is None:
            self.version()
        return self._bits

    def nonce(self):
        if self._version is None:
            self.version()
        return self._nonce

    def txn_count(self):
        if self._version is None:
            self.version()
        return self._txn_count

    def header(self):
        if self._txns is None:
            if self._tx_offset is None:
                self.version()
            self._header = self._memoryview[0:self._tx_offset]
        return self._header

    def txns(self):
        if self._txns is None:
            if self._tx_offset is None:
                self.version()

            self._txns = list()

            start = self._tx_offset
            for _ in range(self.txn_count()):
                size = get_next_tx_size(self.buf, start)
                self._txns.append(self._memoryview[start:start + size])
                start += size

        return self._txns

    def block_hash(self) -> BtcObjectHash:
        if self._hash_val is None:
            header = self._memoryview[BTC_HDR_COMMON_OFF:BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE]
            raw_hash = crypto.bitcoin_hash(header)
            self._hash_val = BtcObjectHash(buf=raw_hash, length=BTC_SHA_HASH_LEN)
        # pyre-fixme[7]: Expected `BtcObjectHash` but got `None`.
        return self._hash_val

    @staticmethod
    def peek_block(input_buffer):
        buf = input_buffer.peek_message(BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE)
        is_here = False
        if input_buffer.length < BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE:
            return is_here, None, None

        header = buf[BTC_HDR_COMMON_OFF:BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE]
        raw_hash = crypto.bitcoin_hash(header)
        payload_len = struct.unpack_from('<L', buf, 16)[0]
        is_here = True
        return is_here, BtcObjectHash(buf=raw_hash, length=BTC_SHA_HASH_LEN), payload_len

    def __repr__(self):
        return "BlockBtcMessage<block_hash: {}, length: {}>".format(self.block_hash(), len(self.rawbytes()))



from argparse import Namespace
from typing import cast

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.models.node_type import NodeType
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.test_utils.helpers import COOKIE_FILE_PATH, get_common_opts, \
    BTC_COMPACT_BLOCK_DECOMPRESS_MIN_TX_COUNT
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway import argument_parsers
from bxgateway.gateway_opts import GatewayOpts


# pylint: disable=unused-argument,too-many-branches
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo


def get_gateway_opts(
    port,
    node_id=None,
    external_ip=constants.LOCALHOST,
    blockchain_address=None,
    test_mode=None,
    peer_gateways=None,
    peer_relays=None,
    peer_transaction_relays=None,
    protocol_version=1,
    bloxroute_version="bloxroute 1.5",
    pub_key=None,
    include_default_ont_args=False,
    blockchain_network_num=constants.DEFAULT_NETWORK_NUM,
    min_peer_gateways=0,
    remote_blockchain_ip=None,
    remote_blockchain_port=None,
    connect_to_remote_blockchain=False,
    blockchain_peers=None,
    blockchain_peers_file=None,
    is_internal_gateway=False,
    is_gateway_miner=False,
    enable_buffered_send=False,
    encrypt_blocks=True,
    cookie_file_path=COOKIE_FILE_PATH,
    blockchain_block_hold_timeout_s=30,
    blockchain_block_recovery_timeout_s=30,
    stay_alive_duration=30 * 60,
    source_version="v1.1.1.1",
    initial_liveliness_check=30,
    block_interval=600,
    non_ssl_port: int = 9001,
    has_fully_updated_tx_service: bool = False,
    max_block_interval_s: float = 10,
    default_tx_flag: TransactionFlag = TransactionFlag.NO_FLAGS,
    log_level_overrides=None,
    enable_network_content_logs=False,
    account_model=None,
    ipc=False,
    ipc_file="bxgateway.ipc",
    ws=False,
    ws_host=constants.LOCALHOST,
    ws_port=28333,
    request_remote_transaction_streaming: bool = False,
    enable_block_compression: bool = True,
    filter_txs_factor: float = 0,
    blockchain_protocol: str = "Ethereum",
    should_restart_on_high_memory: bool = False,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID,
    **kwargs,
) -> GatewayOpts:
    if node_id is None:
        node_id = "Gateway at {0}".format(port)
    if peer_gateways is None:
        peer_gateways = set()
    if peer_relays is None:
        peer_relays = set()
    if peer_transaction_relays is None:
        peer_transaction_relays = []
    if blockchain_address is None:
        blockchain_address = ("127.0.0.1", 7000)  # not real, just a placeholder
    if test_mode is None:
        test_mode = []
    if log_level_overrides is None:
        log_level_overrides = {}
    if remote_blockchain_ip is not None and remote_blockchain_port is not None:
        remote_blockchain_peer = (remote_blockchain_ip, remote_blockchain_port)
    else:
        remote_blockchain_peer = None
    # below is the same as `class ParseBlockchainPeers(argparse.Action)`
    if blockchain_peers is not None:
        blockchain_peers_set = set()
        blockchain_protocol = blockchain_protocol
        for peer in blockchain_peers.split(","):
            blockchain_peer = argument_parsers.parse_peer(blockchain_protocol, peer) # from bxgateway import argument_parsers,
            blockchain_peers_set.add(blockchain_peer)
        blockchain_peers = blockchain_peers_set

    partial_apply_args = locals().copy()
    for kwarg, arg in partial_apply_args["kwargs"].items():
        partial_apply_args[kwarg] = arg

    partial_apply_args["outbound_peers"] = set(peer_gateways).union(peer_relays)

    opts = Namespace()
    common_opts = get_common_opts(**partial_apply_args)
    opts.__dict__.update(common_opts.__dict__)

    opts.__dict__.update(
        {
            "bloxroute_version": bloxroute_version,
            "blockchain_ip": blockchain_address[0],
            "blockchain_port": blockchain_address[1],
            "blockchain_protocol": blockchain_protocol,
            "blockchain_network": "Mainnet",
            "blockchain_version": 12345,
            "test_mode": test_mode,
            "peer_gateways": peer_gateways,
            "peer_relays": peer_relays,
            "peer_transaction_relays": peer_transaction_relays,
            "protocol_version": protocol_version,
            "min_peer_gateways": min_peer_gateways,
            "remote_blockchain_ip": remote_blockchain_ip,
            "remote_blockchain_port": remote_blockchain_port,
            "remote_blockchain_peer": remote_blockchain_peer,
            "connect_to_remote_blockchain": connect_to_remote_blockchain,
            "blockchain_peers": blockchain_peers,
            "blockchain_peers_file": blockchain_peers_file,
            "is_internal_gateway": is_internal_gateway,
            "is_gateway_miner": is_gateway_miner,
            "encrypt_blocks": encrypt_blocks,
            "enable_buffered_send": enable_buffered_send,
            "compact_block": True,
            "compact_block_min_tx_count": BTC_COMPACT_BLOCK_DECOMPRESS_MIN_TX_COUNT,
            "tune_send_buffer_size": False,
            "dump_short_id_mapping_compression": False,
            "max_block_interval_s": max_block_interval_s,
            "cookie_file_path": cookie_file_path,
            "config_update_interval": 60,
            "blockchain_message_ttl": 10,
            "remote_blockchain_message_ttl": 10,
            "stay_alive_duration": stay_alive_duration,
            "initial_liveliness_check": initial_liveliness_check,
            "has_fully_updated_tx_service": has_fully_updated_tx_service,
            "source_version": source_version,
            "require_blockchain_connection": True,
            "non_ssl_port": non_ssl_port,
            "default_tx_flag": default_tx_flag,
            "should_update_source_version": False,
            "enable_network_content_logs": False,
            "enable_node_cache": True,
            "dump_short_id_mapping_compression_path": "",
            "ws": ws,
            "ws_host": constants.LOCALHOST,
            "ws_port": 28333,
            "account_id": account_id,
            "ipc": False,
            "ipc_file": "bxgateway.ipc",
            "request_remote_transaction_streaming": request_remote_transaction_streaming,
            "process_node_txs_in_extension": True,
            "enable_eth_extensions": True,   # TODO remove,
            "request_recovery": True,
            "enable_block_compression": enable_block_compression,
            "filter_txs_factor": filter_txs_factor,
            "min_peer_relays_count": None,
            "should_restart_on_high_memory": should_restart_on_high_memory,
            "stream_to_peer_gateway": None,
        }
    )

    # bitcoin
    opts.__dict__.update(
        {
            "blockchain_net_magic": 12345,
            "blockchain_version": 23456,
            "blockchain_nonce": 0,
            "blockchain_services": 1,
        }
    )
    # ethereum
    opts.__dict__.update(
        {
            "private_key": "294549f8629f0eeb2b8e01aca491f701f5386a9662403b485c4efe7d447dfba3",
            "node_public_key": pub_key,
            "remote_public_key": pub_key,
            "network_id": 1,
            "chain_difficulty": 4194304,
            "genesis_hash": "1e8ff5fd9d06ab673db775cf5c72a6b2d63171cd26fe1e6a8b9d2d696049c781",
            "no_discovery": True,
            "enode": "enode://294549f8629f0eeb2b8e01aca491f701f5386a9662403b485c4efe7d447dfba3@127.0.0.1:8000",
            "eth_ws_uri": None
        }
    )
    # ontology
    opts.__dict__.update(
        {
            "blockchain_net_magic": 12345,
            "blockchain_version": 23456,
            "is_consensus": True,
            "sync_port": 10001,
            "http_info_port": 10002,
            "consensus_port": 10003,
            "cap": bytes(32),
            "blockchain_nonce": 0,
            "relay": True,
            "soft_version": "myversion",
            "blockchain_services": 1,
        }
    )
    for key, val in kwargs.items():
        opts.__dict__[key] = val

    gateway_opts = GatewayOpts.from_opts(opts)
    if account_model:
        gateway_opts.set_account_options(account_model)

    # some attributes are usually set by the node runner
    gateway_opts.__dict__.update({
        "node_type": NodeType.EXTERNAL_GATEWAY,
        "outbound_peers": common_opts.outbound_peers,
        "sid_expire_time": common_opts.sid_expire_time,
        "blockchain_networks": common_opts.blockchain_networks,
        "blockchain_network_num": common_opts.blockchain_network_num,
        "split_relays": common_opts.split_relays,
        "should_update_source_version": False,
        "blockchain_block_interval": block_interval,
        "blockchain_ignore_block_interval_count": 3,
        "blockchain_block_recovery_timeout_s": blockchain_block_recovery_timeout_s,
        "blockchain_block_hold_timeout_s": blockchain_block_hold_timeout_s,
    })
    return gateway_opts


def add_blockchain_peer(node: AbstractGatewayNode, connection: AbstractConnection):
    node_conn = cast(AbstractGatewayBlockchainConnection, connection)
    node.blockchain_peers.add(BlockchainPeerInfo(node_conn.peer_ip, node_conn.peer_port))
    block_queuing_service = node.build_block_queuing_service(node_conn)
    node.block_queuing_service_manager.add_block_queuing_service(node_conn, block_queuing_service)



import argparse
import asyncio
import sys
import time
from datetime import datetime
from typing import Optional, Dict, TextIO, Set, cast

from bloxroute_cli.provider.cloud_wss_provider import CloudWssProvider
from bloxroute_cli.provider.ws_provider import WsProvider
from bxcommon.rpc.external.eth_ws_subscriber import EthWsSubscriber


class HashEntry:
    block_hash = None
    gateway_time_received: Optional[float] = None
    eth_node_time_received: Optional[float] = None

    def __init__(
        self, block_hash: str, gateway_time_received: Optional[float] = None, node_time_received: Optional[float] = None
    ) -> None:
        self.block_hash = block_hash
        self.gateway_time_received = gateway_time_received
        self.eth_node_time_received = node_time_received

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, HashEntry)
            and other.block_hash == self.block_hash
        )


ws_provider: WsProvider
eth_ws_provider: EthWsSubscriber
time_to_begin_comparison = 0.0
time_to_end_comparison = 0.0
all_hashes_file: TextIO
missing_hashes_file: TextIO
trail_new_hashes: Set[str] = set()
lead_new_hashes: Set[str] = set()
seen_hashes: Dict[str, HashEntry] = {}
num_intervals_passed = 0
num_intervals: Optional[int] = None


async def main() -> None:
    global seen_hashes
    global trail_new_hashes
    global time_to_begin_comparison
    global time_to_end_comparison
    global num_intervals_passed
    global num_intervals

    parser = get_argument_parser()
    args = parser.parse_args()

    global all_hashes_file
    if "ALL" in args.dump:
        all_hashes_file = cast(TextIO, open("all_block_hashes.csv", "w"))
        all_hashes_file.write("block-hash,blxr Time,eth Time\n")
    else:
        all_hashes_file = cast(TextIO, None)

    global missing_hashes_file
    if "MISSING" in args.dump:
        missing_hashes_file = cast(TextIO, open("missing_block_hashes.txt", "w"))
    else:
        missing_hashes_file = cast(TextIO, None)

    time_to_begin_comparison = time.time() + args.lead_time
    time_to_end_comparison = time_to_begin_comparison + args.interval
    num_intervals = args.num_intervals

    if args.use_cloud_api:
        asyncio.create_task(process_new_blocks_cloud_api(
            args.cloud_api_ws_uri,
            args.auth_header,
            args.feed_name,
            args.exclude_block_contents
        ))
    else:
        asyncio.create_task(process_new_blocks_gateway(
            args.gateway,
            args.feed_name,
            args.exclude_block_contents
        ))
    asyncio.create_task(process_new_blocks_eth(
        args.eth,
        args.exclude_block_contents
    ))

    await asyncio.sleep(args.lead_time)
    while True:
        await asyncio.sleep(args.interval)
        trail_new_hashes.clear()
        await asyncio.sleep(args.trail_time)
        print(
            f"-----------------------------------------------------\n"
            f"Interval: {args.interval} seconds. "
            f"End time: {datetime.now().isoformat(sep=' ', timespec='seconds')} {stats(args.ignore_delta)}"
        )
        seen_hashes.clear()
        lead_new_hashes.clear()
        time_to_end_comparison = time.time() + args.interval

        num_intervals_passed += 1
        if num_intervals_passed == args.num_intervals:
            print(f"{num_intervals_passed} of {num_intervals} intervals complete. Exiting.")
            sys.exit(0)


async def process_new_blocks_bdn(feed_name: str, exclude_block_contents: bool):
    options = None
    if exclude_block_contents:
        options = {"include": ["hash", "header"]}

    subscription_id = await ws_provider.subscribe(feed_name, options)

    while True:
        next_notification = await ws_provider.get_next_subscription_notification_by_id(subscription_id)
        time_received = time.time()
        block_hash = next_notification.notification["hash"]

        if time_received < time_to_begin_comparison:
            lead_new_hashes.add(block_hash)
            continue

        if block_hash in seen_hashes:
            hash_entry = seen_hashes[block_hash]
            if hash_entry.gateway_time_received is None:
                hash_entry.gateway_time_received = time_received
        elif time_received < time_to_end_comparison and block_hash not in trail_new_hashes and block_hash not in lead_new_hashes:
            seen_hashes[block_hash] = HashEntry(block_hash, gateway_time_received=time.time())
        else:
            trail_new_hashes.add(block_hash)


async def process_new_blocks_cloud_api(
    ws_uri: str, auth_header: str, feed_name: str, exclude_block_contents: bool
) -> None:
    print(f"Initiating connection to: {ws_uri}")
    async with WsProvider(
        uri=ws_uri,
        headers={"Authorization": auth_header}
    ) as ws:
        print(f"websockets endpoint: {ws_uri} established")
        global ws_provider
        ws_provider = cast(CloudWssProvider, ws)

        await process_new_blocks_bdn(feed_name, exclude_block_contents)


async def process_new_blocks_gateway(gateway_url: str, feed_name: str, exclude_block_contents: bool) -> None:
    print(f"Initiating connection to: {gateway_url}")
    async with WsProvider(gateway_url) as ws:
        print(f"websockets endpoint: {gateway_url} established")
        global ws_provider
        ws_provider = cast(WsProvider, ws)

        await process_new_blocks_bdn(feed_name, exclude_block_contents)


def handle_eth_block_hash(block_hash, time_received: float):
    if block_hash in seen_hashes:
        hash_entry = seen_hashes[block_hash]
        if hash_entry.eth_node_time_received is None:
            hash_entry.eth_node_time_received = time_received
    elif time_received < time_to_end_comparison and block_hash not in trail_new_hashes and block_hash not in lead_new_hashes:
        seen_hashes[block_hash] = HashEntry(block_hash, node_time_received=time_received)
    else:
        trail_new_hashes.add(block_hash)


async def process_new_blocks_eth(eth_url: str, exclude_block_contents: bool) -> None:
    print(f"Initiating connection to: {eth_url}")
    async with EthWsSubscriber(eth_url) as eth_ws:
        print(f"websockets endpoint: {eth_url} established")

        global eth_ws_provider
        eth_ws_provider = cast(EthWsSubscriber, eth_ws)
        subscription_id = await eth_ws.subscribe("newHeads")

        while True:
            next_notification = await eth_ws.get_next_subscription_notification_by_id(subscription_id)
            time_received = time.time()

            header = next_notification.notification
            block_hash = header["hash"]

            if time_received < time_to_begin_comparison:
                lead_new_hashes.add(block_hash)
                continue

            if not exclude_block_contents:
                asyncio.create_task(fetch_block(block_hash))
            else:
                handle_eth_block_hash(block_hash, time_received)


async def fetch_block(block_hash: str) -> None:
    try:
        response = await eth_ws_provider.call_rpc(
            "eth_getBlockByHash",
            [block_hash, True]
        )
        time_received = time.time()
        block = response.result
        if block is None:
            return

        handle_eth_block_hash(block_hash, time_received)

    except Exception:
        print(f"Attempt to fetch full block for {block_hash} failed. Abandoning.")
        pass


def stats(ignore_delta: int) -> str:
    blocks_seen_by_both_feeds_gateway_first = 0
    blocks_seen_by_both_feeds_eth_node_first = 0
    block_received_by_gateway_first_total_delta = 0
    block_received_by_eth_node_first_total_delta = 0
    new_blocks_from_gateway_feed_first = 0
    new_blocks_from_eth_node_feed_first = 0
    total_blocks_from_gateway = 0
    total_blocks_from_eth_node = 0

    for block_hash, hash_entry in seen_hashes.items():
        if hash_entry.gateway_time_received is None:
            eth_node_time_received = hash_entry.eth_node_time_received
            assert isinstance(eth_node_time_received, float)

            if missing_hashes_file:
                missing_hashes_file.write(f"{block_hash}\n")
            if all_hashes_file:
                all_hashes_file.write(f"{block_hash}, 0, {datetime.fromtimestamp(eth_node_time_received)}\n")
            new_blocks_from_eth_node_feed_first += 1
            total_blocks_from_eth_node += 1
            continue
        if hash_entry.eth_node_time_received is None:
            gateway_time_received = hash_entry.gateway_time_received
            assert isinstance(gateway_time_received, float)

            if all_hashes_file:
                all_hashes_file.write(f"{block_hash}, {datetime.fromtimestamp(gateway_time_received)}, 0\n")
            new_blocks_from_gateway_feed_first += 1
            total_blocks_from_gateway += 1
            continue

        eth_node_time_received = hash_entry.eth_node_time_received
        gateway_time_received = hash_entry.gateway_time_received
        assert isinstance(eth_node_time_received, float)
        assert isinstance(gateway_time_received, float)

        total_blocks_from_gateway += 1
        total_blocks_from_eth_node += 1

        if all_hashes_file:
            all_hashes_file.write(
                f"{block_hash},"
                f"{datetime.fromtimestamp(gateway_time_received)},"
                f"{datetime.fromtimestamp(eth_node_time_received)}\n"
            )

        if abs(gateway_time_received - eth_node_time_received) > ignore_delta:
            continue

        if gateway_time_received < eth_node_time_received:
            new_blocks_from_gateway_feed_first += 1
            blocks_seen_by_both_feeds_gateway_first += 1
            block_received_by_gateway_first_total_delta += eth_node_time_received - gateway_time_received
        elif eth_node_time_received < gateway_time_received:
            new_blocks_from_eth_node_feed_first += 1
            blocks_seen_by_both_feeds_eth_node_first += 1
            block_received_by_eth_node_first_total_delta += gateway_time_received - eth_node_time_received

    new_block_seen_by_both_feeds = blocks_seen_by_both_feeds_gateway_first + blocks_seen_by_both_feeds_eth_node_first
    block_received_by_gw_first_avg_delta = 0 if blocks_seen_by_both_feeds_gateway_first == 0 \
        else round((block_received_by_gateway_first_total_delta / float(blocks_seen_by_both_feeds_gateway_first) * 1000), 1)
    block_received_by_eth_node_first_avg_delta = 0 if blocks_seen_by_both_feeds_eth_node_first == 0 \
        else round((block_received_by_eth_node_first_total_delta / float(blocks_seen_by_both_feeds_eth_node_first) * 1000), 1)
    percentage_block_seen_first_by_gateway = 0 if new_block_seen_by_both_feeds == 0 \
        else int((blocks_seen_by_both_feeds_gateway_first / new_block_seen_by_both_feeds) * 100)

    return (
        f"\nBlocks summary:\n"
        f"Number of new blocks received first from gateway: {new_blocks_from_gateway_feed_first}\n"
        f"Number of new blocks received first from node: {new_blocks_from_eth_node_feed_first}\n"
        f"Total number of blocks seen: {new_blocks_from_eth_node_feed_first + new_blocks_from_gateway_feed_first}\n"
        f"Total blocks from gateway: {total_blocks_from_gateway}\n"
        f"Total blocks from eth node: {total_blocks_from_eth_node}\n"

        f"\nAnalysis of Blocks received on both feeds:\n"
        f"Number of blocks: {new_block_seen_by_both_feeds}\n"
        f"Number of blocks received from Gateway first: {blocks_seen_by_both_feeds_gateway_first}\n"
        f"Number of blocks received from Ethereum node first: {blocks_seen_by_both_feeds_eth_node_first}\n"
        f"Percentage of blocks seen first from gateway: {percentage_block_seen_first_by_gateway}%\n"
        f"Average time difference for blocks received first from gateway (ms): {block_received_by_gw_first_avg_delta}\n"
        f"Average time difference for blocks received first from Ethereum node (ms): {block_received_by_eth_node_first_avg_delta}\n"
    )


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--gateway", type=str, default="ws://127.0.0.1:28333")
    parser.add_argument("--eth", type=str, default="ws://127.0.0.1:8546")
    parser.add_argument("--feed-name", type=str, default="newBlocks", choices=["newBlocks"])
    parser.add_argument("--exclude-block-contents", action="store_true")
    parser.add_argument("--interval", type=int, default=600, help="Length of feed sample interval in seconds")
    parser.add_argument("--num-intervals", type=int, default=6)
    parser.add_argument("--lead-time", type=int, default=60, help="Seconds to wait before starting to compare feeds")
    parser.add_argument("--trail-time", type=int, default=60, help="Seconds to wait after interval to receive block on both feeds")
    parser.add_argument("--dump", type=str, default="", choices=["ALL", "MISSING", "ALL,MISSING"])
    parser.add_argument("--ignore-delta", type=int, default=5, help="Ignore block with delta above this amount (seconds)")
    parser.add_argument("--use-cloud-api", action="store_true")
    parser.add_argument("--auth-header", type=str, help="Authorization header created with account id and password")
    parser.add_argument("--cloud-api-ws-uri", type=str, default="wss://api.blxrbdn.com/ws")
    parser.add_argument("--verbose", action="store_true")
    return parser


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(main())
    except KeyboardInterrupt:
        print(f"{num_intervals_passed} of {num_intervals} intervals complete. Exiting.")
        sys.exit(0)



import logging
from enum import IntEnum


class LogLevel(IntEnum):
    CRITICAL = 50
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    STATS = 25
    INFO = 20
    DEBUG = 10
    TRACE = 5
    NOTSET = 0


logging.addLevelName(LogLevel.STATS, "STATS")
logging.addLevelName(LogLevel.TRACE, "TRACE")


def from_string(level: str) -> LogLevel:
    return LogLevel[level.upper()]



from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


class VerAckBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.VERACK

    def __init__(self, magic=None, buf=None):
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF)
            self.buf = buf

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, 0, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None



from bxgateway.messages.gateway.abstract_blockchain_sync_message import AbstractBlockchainSyncMessage
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class BlockchainSyncResponseMessage(AbstractBlockchainSyncMessage):
    """
    Message type for receiving direct blockchain messages for syncing chainstate.

    NOTE:
    This class is currently unused. See `blockchain_sync_service.py` for reasons.
    """
    MESSAGE_TYPE = GatewayMessageType.SYNC_RESPONSE



from typing import Optional

from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class NotFoundOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.NOT_FOUND

    def __init__(self, magic: Optional[int] = None, block_hash: Optional[OntObjectHash] = None,
                 buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_HASH_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            assert block_hash is not None
            buf[off:off + ont_constants.ONT_HASH_LEN] = block_hash.get_little_endian()
            off += ont_constants.ONT_HASH_LEN

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._block_hash = None



import typing

from bxcommon.messages.bloxroute.compact_block_short_ids_serializer import BlockOffsets
from bxcommon.utils.object_hash import Sha256Hash


class BlockHeaderInfo(typing.NamedTuple):
    block_offsets: BlockOffsets
    short_ids: typing.List[int]
    short_ids_len: int
    block_hash: Sha256Hash
    offset: int
    txn_count: int



from abc import abstractmethod
from typing import Generic, Union, Dict, Any, List

from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest, Node
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError
from bxcommon.utils import convert


class AbstractBlxrTransactionRpcRequest(AbstractRpcRequest, Generic[Node]):
    SYNCHRONOUS = rpc_constants.SYNCHRONOUS_PARAMS_KEY
    track_flag: TransactionFlag = TransactionFlag.PAID_TX

    help = {
        "params": f"[Required - {rpc_constants.TRANSACTION_PARAMS_KEY}: [transaction payload in hex string format]]. "
        f"Optional - {SYNCHRONOUS}: [True (wait for response from the relay - default), "
        "False (don't wait for response)]."
        f"{rpc_constants.BLOCKCHAIN_PROTOCOL_PARAMS_KEY}: [by default: `Ethereum`] "
        f"{rpc_constants.BLOCKCHAIN_NETWORK_PARAMS_KEY}: [by default: `Mainnet`], ",
        "description": "send transaction to the bloXroute BDN"
    }

    def validate_params(self) -> None:
        params = self.params
        if params is None or not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )
        assert params is not None
        self.validate_transaction_param(params)

    def validate_transaction_param(self, params: Union[Dict[str, Any], List[Any], None]) -> None:
        assert params is not None
        assert isinstance(params, dict)

        if rpc_constants.TRANSACTION_PARAMS_KEY not in params:
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid transaction request params type: {self.params}"
            )

    def parse_track_flag(self, params: Union[Dict[str, Any], List[Any], None]):
        assert params is not None
        assert isinstance(params, dict)

        if not self.track_flag:
            self.track_flag = TransactionFlag.PAID_TX

        if rpc_constants.STATUS_TRACK_PARAMS_KEY in params:
            track_flag_str = params[rpc_constants.STATUS_TRACK_PARAMS_KEY]
            track_flag = convert.str_to_bool(str(track_flag_str).lower(), default=False)
            if track_flag:
                self.track_flag |= TransactionFlag.STATUS_TRACK

        if rpc_constants.NONCE_MONITORING_PARAMS_KEY in params:
            nonce_flag_str = params[rpc_constants.NONCE_MONITORING_PARAMS_KEY]
            nonce_flag = convert.str_to_bool(str(nonce_flag_str).lower(), default=False)
            if nonce_flag:
                self.track_flag |= TransactionFlag.NONCE_TRACK

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, dict)

        account_id = self.get_account_id()
        if not account_id:
            raise RpcAccountIdError(
                self.request_id,
                "Account ID is missing."
            )

        transaction_str: str = params[rpc_constants.TRANSACTION_PARAMS_KEY]
        network_num = self.get_network_num()
        self.parse_track_flag(params)

        return await self.process_transaction(network_num, account_id, transaction_str)

    @abstractmethod
    async def process_transaction(
        self, network_num: int, account_id: str, transaction_str: str
    ) -> JsonRpcResponse:
        pass

    @abstractmethod
    def get_network_num(self) -> int:
        pass

    @abstractmethod
    def get_account_id(self) -> str:
        pass



import os
import functools
from json import JSONDecodeError
from collections import defaultdict
from typing import TYPE_CHECKING

from bxutils import logging
from bxutils.logging import log_config
from bxutils.logging import log_level
from bxcommon.utils import model_loader
from bxcommon.utils import config

from bxgateway import log_messages
from bxgateway.models.config.gateway_node_config_model import GatewayNodeConfigModel
from bxgateway import gateway_constants

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)
project_root_logger = logging.get_logger(__name__.split(".")[0])

config_file_last_updated = defaultdict(float)
last_config_model = defaultdict(GatewayNodeConfigModel)


def read_config_file(full_path: str) -> GatewayNodeConfigModel:
    try:
        file_mod_time = os.stat(full_path).st_mtime
    except FileNotFoundError:
        logger.debug("Config file not found {}", full_path)
        return GatewayNodeConfigModel()
    if config_file_last_updated.get(full_path) != file_mod_time:
        try:
            with open(full_path, "r") as f:
                raw_input = f.read()
                config_model = model_loader.load_model_from_json(
                    model_class=GatewayNodeConfigModel, model_params=raw_input
                )
                config_file_last_updated[full_path] = file_mod_time
                last_config_model[full_path] = config_model
        except (TypeError, JSONDecodeError):
            logger.warning(log_messages.INVALID_CONFIG_PAYLOAD, full_path)
            return GatewayNodeConfigModel()
        except Exception as e:
            logger.warning(log_messages.READ_CONFIG_FAIL, full_path, e)
            return GatewayNodeConfigModel()
        return config_model
    else:
        return last_config_model[full_path]


def compare_and_update(new_value, target, setter, item=None):
    if new_value is not None:
        if new_value != target:
            setter(new_value)
            logger.trace("Updating item: {}. New value: {}, old value: {}.", item, repr(new_value), repr(target))
            return


def update_node_config(node: "AbstractGatewayNode"):
    config.init_file_in_data_dir(gateway_constants.CONFIG_FILE_NAME)
    config.init_file_in_data_dir(gateway_constants.CONFIG_OVERRIDE_NAME)
    default_node_config = read_config_file(config.get_data_file(gateway_constants.CONFIG_FILE_NAME))
    override_node_config = read_config_file(config.get_data_file(gateway_constants.CONFIG_OVERRIDE_NAME))
    node_config = GatewayNodeConfigModel()
    node_config.merge(default_node_config)
    node_config.merge(override_node_config)

    logger.trace({"type": "update_node_config", "data": node_config})
    log_config_model = node_config.log_config
    if log_config_model is not None:
        try:
            if log_config_model.log_level is not None:
                # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[str]`.
                compare_and_update(log_level.from_string(log_config_model.log_level),
                                   project_root_logger.level,
                                   setter=functools.partial(log_config.set_level, node.opts.logger_names),
                                   item="log_level")
            if log_config_model.log_level_overrides is not None:
                # pyre-fixme[6]: Expected `Dict[str,
                #  typing.Union[log_level.LogLevel, str]]` for 1st param but got
                #  `Dict[str, str]`.
                log_config.set_log_levels(log_config_model.log_level_overrides)
        except (AttributeError, KeyError):
            logger.warning(log_messages.INVALID_CONFIG_LOG_LEVEL, log_config_model.log_level)

    if node_config.cron_config is not None:
        compare_and_update(node_config.cron_config.config_update_interval,
                           node.opts.config_update_interval,
                           item="config_update_interval",
                           setter=lambda val: node.opts.__setattr__("config_update_interval", val)
                           )
        compare_and_update(node_config.cron_config.info_stats_interval,
                           node.opts.info_stats_interval,
                           item="info_stats_interval",
                           setter=lambda val: node.opts.__setattr__("info_stats_interval", val))
        compare_and_update(node_config.cron_config.throughput_stats_interval,
                           node.opts.throughput_stats_interval,
                           item="throughput_stats_interval",
                           setter=lambda val: node.opts.__setattr__("throughput_stats_interval", val))
        compare_and_update(node_config.cron_config.memory_stats_interval,
                           node.opts.memory_stats_interval,
                           item="memory_stats_interval",
                           setter=lambda val: node.opts.__setattr__("memory_stats_interval", val))



import struct
import time

from bxgateway.btc_constants import BTC_NODE_SERVICES, BTC_HDR_COMMON_OFF, BTC_IP_ADDR_PORT_SIZE
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import btcbytearray_to_ipaddrport, ipaddrport_to_btcbytearray


class VersionBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.VERSION

    def __init__(self, magic=None, version=None, dst_ip=None, dst_port=None,
                 src_ip=None, src_port=None, nonce=None, start_height=None, user_agent=None, services=BTC_NODE_SERVICES,
                 buf=None):

        if buf is None:
            buf = bytearray(109 + len(user_agent))
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            struct.pack_into('<IQQQ', buf, off, version, services, int(time.time()), 1)
            off += 28
            buf[off:off + BTC_IP_ADDR_PORT_SIZE] = ipaddrport_to_btcbytearray(dst_ip, dst_port)
            off += BTC_IP_ADDR_PORT_SIZE
            struct.pack_into('<Q', buf, off, 1)
            off += 8
            buf[off:off + BTC_IP_ADDR_PORT_SIZE] = ipaddrport_to_btcbytearray(src_ip, src_port)
            off += BTC_IP_ADDR_PORT_SIZE
            struct.pack_into('<Q%dpL' % (len(user_agent) + 1,), buf, off, nonce, user_agent, start_height)
            off += 8 + len(user_agent) + 1 + 4
            # we do not send or parse the relay boolean. if present, we benignly ignore it.

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._memoryview = memoryview(buf)
        self._version = self._services = self._timestamp = self._dst_ip = self._dst_port = None
        self._src_ip = self._src_port = self._nonce = self._start_height = self._user_agent = None

    def version(self):
        if self._version is None:
            off = BTC_HDR_COMMON_OFF
            self._version, self._services, self._timestamp = struct.unpack_from('<IQQ', self.buf, off)

        return self._version

    def services(self):
        if self._services is None:
            self.version()

        return self._services

    def timestamp(self):
        if self._version is None:
            self.version()
        return self._timestamp

    def dst_ip(self):
        if self._dst_ip is None:
            dst_ip_off = BTC_HDR_COMMON_OFF + 28
            self._dst_ip, self._dst_port = btcbytearray_to_ipaddrport(self.buf[dst_ip_off:dst_ip_off + BTC_IP_ADDR_PORT_SIZE])
        return self._dst_ip

    def dst_port(self):
        if self._dst_port is None:
            self.dst_ip()
        return self._dst_port

    def src_ip(self):
        if self._src_ip is None:
            src_ip_off = BTC_HDR_COMMON_OFF + 54
            self._src_ip, self._src_port = btcbytearray_to_ipaddrport(self.buf[src_ip_off:src_ip_off + BTC_IP_ADDR_PORT_SIZE])
        return self._src_ip

    def src_port(self):
        if self._src_port is None:
            self.src_ip()
        return self._src_port

    def nonce(self):
        if self._nonce is None:
            nonce_off = BTC_HDR_COMMON_OFF + 72
            self._nonce = struct.unpack_from('<Q', self.buf, nonce_off)[0]
        return self._nonce

    def user_agent(self):
        if self._user_agent is None:
            user_agent_off = BTC_HDR_COMMON_OFF + 80
            self._user_agent = struct.unpack_from('%dp' % (len(self.buf) - user_agent_off,), self.buf, user_agent_off)[
                0]
            height_off = user_agent_off + len(self._user_agent) + 1
            self._start_height = struct.unpack_from('<L', self.buf, height_off)[0]
        return self._user_agent

    def start_height(self):
        if self._user_agent is None:
            self.user_agent()
        return self._start_height



from dataclasses import dataclass
from typing import Optional

from bxgateway.utils.logging.status.connection_info import ConnectionInfo


@dataclass
class BlockchainConnection(ConnectionInfo):

    def __init__(self, ip_address: str, port: str, fileno: Optional[str] = None, connection_time: Optional[str] = None):
        self.ip_address = ip_address
        self.port = port
        self.fileno = fileno
        self.connection_time = connection_time

    def __eq__(self, other: "BlockchainConnection") -> bool:
        return self.ip_address == other.ip_address \
               and self.port == other.port


import typing
import time
from typing import TYPE_CHECKING, Optional, Set
from datetime import datetime

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.messages.bloxroute.block_confirmation_message import BlockConfirmationMessage
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.proxy.task_queue_proxy import TaskQueueProxy
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.memory_utils import SpecialTuple
from bxcommon.utils import memory_utils
from bxcommon.services import extension_cleanup_service_helpers

from bxgateway import btc_constants
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.services.btc.abstract_btc_block_cleanup_service import AbstractBtcBlockCleanupService

import task_pool_executor as tpe

if TYPE_CHECKING:
    from bxgateway.connections.btc.btc_gateway_node import BtcGatewayNode

logger = logging.get_logger(LogRecordType.BlockCleanup, __name__)


def create_block_confirmation_cleanup_task() -> tpe.BlockConfirmationCleanupTask:
    return tpe.BlockConfirmationCleanupTask()


class BtcExtensionBlockCleanupService(AbstractBtcBlockCleanupService):

    MINIMAL_SUB_TASK_TX_COUNT = btc_constants.BTC_MINIMAL_SUB_TASK_TX_COUNT

    def __init__(self, node: "BtcGatewayNode", network_num: int):
        super(BtcExtensionBlockCleanupService, self).__init__(node, network_num)
        self.block_cleanup_tasks = TaskQueueProxy(self._create_block_cleanup_task)
        self.block_confirmation_cleanup_tasks = TaskQueueProxy(create_block_confirmation_cleanup_task)

    def clean_block_transactions(
            self, block_msg: BlockBtcMessage, transaction_service: TransactionService
    ) -> None:
        start_datetime = datetime.utcnow()
        start_time = time.time()
        tx_hash_to_contents_len_before_cleanup = transaction_service.get_tx_hash_to_contents_len()
        cleanup_task = self.block_cleanup_tasks.borrow_task()
        tx_service = typing.cast(ExtensionTransactionService, transaction_service)
        cleanup_task.init(tpe.InputBytes(block_msg.buf), tx_service.proxy)
        init_time = time.time()
        task_pool_proxy.run_task(cleanup_task)
        task_run_time = time.time()
        unknown_tx_hashes_count = len(cleanup_task.unknown_tx_hashes())
        tx_property_fetch_time = time.time()
        short_ids = cleanup_task.short_ids()
        short_ids_fetch_time = time.time()
        short_ids_count = len(short_ids)
        tx_service.update_removed_transactions(cleanup_task.total_content_removed(), short_ids)
        remove_from_tx_service_time = time.time()
        # TODO : clean the short ids/transactions from the alarm queue after refactoring the transaction service
        block_hash = block_msg.block_hash()
        tx_service.on_block_cleaned_up(block_hash)
        tx_hash_to_contents_len_after_cleanup = transaction_service.get_tx_hash_to_contents_len()
        end_datetime = datetime.utcnow()
        end_time = time.time()

        logger.statistics(
            {
                "type": "BlockTransactionsCleanup",
                "block_hash": repr(block_hash),
                "unknown_tx_hashes_count": unknown_tx_hashes_count,
                "short_ids_count": short_ids_count,
                "block_transactions_count": cleanup_task.txn_count(),
                "start_datetime": start_datetime,
                "end_datetime": end_datetime,
                "task_init_time": init_time - start_time,
                "task_run_time": task_run_time - init_time,
                "tx_property_fetch_time": tx_property_fetch_time - task_run_time,
                "short_ids_fetch_time": short_ids_fetch_time - tx_property_fetch_time,
                "remove_from_tx_service_time": remove_from_tx_service_time - short_ids_fetch_time,
                "duration": end_time - start_time,
                "tx_hash_to_contents_len_before_cleanup": tx_hash_to_contents_len_before_cleanup,
                "tx_hash_to_contents_len_after_cleanup": tx_hash_to_contents_len_after_cleanup,
            }
        )
        self.block_cleanup_tasks.return_task(cleanup_task)
        self._block_hash_marked_for_cleanup.discard(block_hash)
        self.node.post_block_cleanup_tasks(
            block_hash=block_hash,
            short_ids=short_ids,
            unknown_tx_hashes=(
                Sha256Hash(convert.hex_to_bytes(tx_hash.hex_string()))
                for tx_hash in cleanup_task.unknown_tx_hashes()
            )
        )

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.add_special_objects(
            self.block_cleanup_tasks, self.block_confirmation_cleanup_tasks, ids
        )

    def _create_block_cleanup_task(self) -> tpe.BtcBlockCleanupTask:
        return tpe.BtcBlockCleanupTask(self.MINIMAL_SUB_TASK_TX_COUNT)

    # pyre-fixme[14]: `contents_cleanup` overrides method defined in
    #  `AbstractBlockCleanupService` inconsistently.
    def contents_cleanup(
        self,
        transaction_service: TransactionService,
        block_confirmation_message: BlockConfirmationMessage
    ):
        extension_cleanup_service_helpers.contents_cleanup(
            transaction_service,
            block_confirmation_message,
            self.block_confirmation_cleanup_tasks
        )



import time
import dataclasses

from dataclasses import dataclass
from typing import Dict, Any, TYPE_CHECKING, List, Type

from bxcommon.feed.feed import FeedKey
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxgateway import gateway_constants
from bxcommon.feed.eth.eth_new_transaction_feed import EthNewTransactionFeed
from bxutils import logging
from bxutils.logging import LogRecordType
from bxcommon.feed.eth.eth_pending_transaction_feed import EthPendingTransactionFeed

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


@dataclass
class TransactionFeedStatInterval(StatsIntervalData):
    new_transaction_received_times: Dict[Sha256Hash, float] = dataclasses.field(default_factory=dict)
    # lower time of the two below
    pending_transaction_received_times: Dict[Sha256Hash, float] = dataclasses.field(default_factory=dict)
    pending_transaction_from_internal_received_times: Dict[Sha256Hash, float] = dataclasses.field(default_factory=dict)
    pending_transaction_from_local_blockchain_received_times: Dict[Sha256Hash, float] = \
        dataclasses.field(default_factory=dict)

    new_transactions_faster_than_pending_times: List[float] = dataclasses.field(default_factory=list)
    pending_from_internal_faster_than_local_times: List[float] = dataclasses.field(default_factory=list)

    pending_transactions_missing_contents: int = 0


class TransactionFeedStatsService(
    StatisticsService[TransactionFeedStatInterval, "AbstractGatewayNode"]
):

    def __init__(
        self,
        interval: int = gateway_constants.GATEWAY_TRANSACTION_FEED_STATS_INTERVAL_S,
        look_back: int = gateway_constants.GATEWAY_TRANSACTION_FEED_STATS_LOOKBACK
    ) -> None:
        super().__init__(
            "TransactionFeedStats",
            interval,
            look_back,
            reset=True,
            stat_logger=logging.get_logger(LogRecordType.TransactionFeedStats)
        )

    def get_interval_data_class(self) -> Type[TransactionFeedStatInterval]:
        return TransactionFeedStatInterval

    def get_info(self) -> Dict[str, Any]:
        interval_data = self.interval_data

        if len(interval_data.new_transactions_faster_than_pending_times) > 0:
            avg_new_transactions_faster_by_s = (
                sum(interval_data.new_transactions_faster_than_pending_times)
                / len(interval_data.new_transactions_faster_than_pending_times)
            )
        else:
            avg_new_transactions_faster_by_s = 0
        new_transactions_faster_count = len([
            diff for diff in interval_data.new_transactions_faster_than_pending_times
            if diff > 0
        ])

        if len(interval_data.pending_from_internal_faster_than_local_times) > 0:
            avg_pending_transactions_from_internal_faster_by_s = (
                sum(interval_data.pending_from_internal_faster_than_local_times)
                / len(interval_data.pending_from_internal_faster_than_local_times)
            )
        else:
            avg_pending_transactions_from_internal_faster_by_s = 0
        pending_transactions_from_internal_faster_count = len([
            diff for diff in interval_data.pending_from_internal_faster_than_local_times
            if diff > 0
        ])

        node = self.node
        assert node is not None
        feeds = node.feed_manager.feeds

        pending_transaction_feed = node.feed_manager.get_feed(FeedKey(EthPendingTransactionFeed.NAME))
        if pending_transaction_feed:
            pending_transaction_feed_subscribers = len(pending_transaction_feed.subscribers)
        else:
            pending_transaction_feed_subscribers = None
        new_transaction_feed = node.feed_manager.get_feed(FeedKey(EthNewTransactionFeed.NAME))
        if new_transaction_feed:
            new_transaction_feed_subscribers = len(new_transaction_feed.subscribers)
        else:
            new_transaction_feed_subscribers = None

        return {
            "start_time": self.interval_data.start_time,
            "end_time": self.interval_data.end_time,
            "new_transactions": len(interval_data.new_transaction_received_times),
            "pending_transactions": len(interval_data.pending_transaction_received_times),
            "pending_transactions_from_internal": len(
                interval_data.pending_transaction_from_internal_received_times
            ),
            "pending_transactions_from_local": len(
                interval_data.pending_transaction_from_local_blockchain_received_times
            ),
            "avg_new_transactions_faster_by_s": avg_new_transactions_faster_by_s,
            "new_transactions_faster_count": new_transactions_faster_count,
            "avg_pending_transactions_from_internal_faster_by_s": avg_pending_transactions_from_internal_faster_by_s,
            "pending_transactions_from_internal_faster_count": pending_transactions_from_internal_faster_count,
            "pending_transactions_missing_contents": interval_data.pending_transactions_missing_contents,
            "pending_transaction_feed_subscribers": pending_transaction_feed_subscribers,
            "new_transaction_feed_subscribers": new_transaction_feed_subscribers,
            "gateway_transaction_streamers": node.get_gateway_transaction_streamers_count()
        }

    def log_new_transaction(self, tx_hash: Sha256Hash) -> None:
        interval_data = self.interval_data
        current_time = time.time()

        interval_data.new_transaction_received_times[tx_hash] = current_time
        if tx_hash in interval_data.pending_transaction_received_times:
            timestamp = interval_data.pending_transaction_received_times[tx_hash]
            interval_data.new_transactions_faster_than_pending_times.append(
                timestamp - current_time
            )

    def log_pending_transaction_from_internal(self, tx_hash: Sha256Hash) -> None:
        interval_data = self.interval_data

        if tx_hash in interval_data.pending_transaction_from_internal_received_times:
            return

        current_time = time.time()

        interval_data.pending_transaction_from_internal_received_times[tx_hash] = current_time
        if tx_hash in interval_data.pending_transaction_received_times:
            # already received from local blockchain
            timestamp = interval_data.pending_transaction_from_local_blockchain_received_times[
                tx_hash
            ]
            interval_data.pending_from_internal_faster_than_local_times.append(
                timestamp - current_time
            )
        else:
            interval_data.pending_transaction_received_times[tx_hash] = current_time
            if tx_hash in interval_data.new_transaction_received_times:
                timestamp = interval_data.new_transaction_received_times[tx_hash]
                interval_data.new_transactions_faster_than_pending_times.append(
                    current_time - timestamp
                )

    def log_pending_transaction_from_local(self, tx_hash: Sha256Hash) -> None:
        interval_data = self.interval_data

        if tx_hash in interval_data.pending_transaction_from_local_blockchain_received_times:
            return

        current_time = time.time()

        interval_data.pending_transaction_from_local_blockchain_received_times[tx_hash] = current_time
        if tx_hash in interval_data.pending_transaction_from_internal_received_times:
            # already received from internal
            timestamp = interval_data.pending_transaction_from_internal_received_times[tx_hash]
            interval_data.pending_from_internal_faster_than_local_times.append(current_time - timestamp)
        else:
            interval_data.pending_transaction_received_times[tx_hash] = current_time
            if tx_hash in interval_data.new_transaction_received_times:
                timestamp = interval_data.new_transaction_received_times[tx_hash]
                interval_data.new_transactions_faster_than_pending_times.append(current_time - timestamp)

    def log_pending_transaction_missing_contents(self) -> None:
        self.interval_data.pending_transactions_missing_contents += 1


transaction_feed_stats_service = TransactionFeedStatsService()



import struct
from typing import Optional

import bxcommon.utils.crypto
from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.models.transaction_info import TransactionInfo
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging
from bxutils.logging.log_level import LogLevel

logger = logging.get_logger(__name__)


class TxContentsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TX_CONTENTS

    """
    Message with tx details. Reply to GetTxContentsMessage.
    """

    def __init__(self, network_num: Optional[int] = None, tx_info: Optional[TransactionInfo] = None,
                 buf: Optional[bytearray] = None):
        """
        Constructor. Expects network number and transaction details, or message bytes.

        :param network_num: blockchain network number
        :param tx: tuple with 3 values (tx short id, tx hash, tx contents)
        :param buf: message bytes
        """
        # msg_size = HDR_COMMON_OFF + network_num + sid + hash + tx size
        msg_size = self.HEADER_LENGTH + constants.NETWORK_NUM_LEN + \
                   constants.SID_LEN + bxcommon.utils.crypto.SHA256_HASH_LEN + constants.UL_INT_SIZE_IN_BYTES + \
                   constants.CONTROL_FLAGS_LEN

        self._network_num = None
        self._tx_info = None

        if buf is None:
            # pyre-fixme[16]: `Optional` has no attribute `contents`.
            tx_contents_len = len(tx_info.contents)
            msg_size += tx_contents_len
            buf = bytearray(msg_size)
            off = self.HEADER_LENGTH

            struct.pack_into("<L", buf, off, network_num)
            off += constants.NETWORK_NUM_LEN

            # pyre-fixme[16]: `Optional` has no attribute `short_id`.
            struct.pack_into("<L", buf, off, tx_info.short_id)
            off += constants.SID_LEN

            # pyre-fixme[16]: `Optional` has no attribute `hash`.
            buf[off:off + bxcommon.utils.crypto.SHA256_HASH_LEN] = tx_info.hash
            off += bxcommon.utils.crypto.SHA256_HASH_LEN

            struct.pack_into("<L", buf, off, tx_contents_len)
            off += constants.UL_INT_SIZE_IN_BYTES

            buf[off:off + tx_contents_len] = tx_info.contents
            off += tx_contents_len

            super(TxContentsMessage, self).__init__(self.MESSAGE_TYPE, msg_size - self.HEADER_LENGTH, buf)
        else:
            if isinstance(buf, str):
                raise TypeError("Buffer can't be string")

            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._payload_len = None
            self._payload = None

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def network_num(self) -> int:
        if self._network_num is None:
            self._parse()

        assert self._network_num is not None
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._network_num

    def get_tx_info(self) -> TransactionInfo:
        if self._tx_info is None:
            self._parse()

        assert self._tx_info is not None
        # pyre-fixme[7]: Expected `TransactionInfo` but got `None`.
        return self._tx_info

    def _parse(self) -> None:
        off = self.HEADER_LENGTH

        self._network_num, = struct.unpack_from("<L", self.buf, off)
        off += constants.NETWORK_NUM_LEN

        tx_sid, = struct.unpack_from("<L", self.buf, off)
        off += constants.SID_LEN

        tx_hash = Sha256Hash(self._memoryview[off:off + bxcommon.utils.crypto.SHA256_HASH_LEN])
        off += bxcommon.utils.crypto.SHA256_HASH_LEN

        tx_size, = struct.unpack_from("<L", self.buf, off)
        off += constants.UL_INT_SIZE_IN_BYTES

        tx = self._memoryview[off:off + tx_size]
        off += tx_size

        self._tx_info = TransactionInfo(tx_hash, tx, tx_sid)

    def __repr__(self) -> str:
        return f"TxContentsMessage<network_num: {self.network_num()}, tx_hash: {self.get_tx_info().hash}, " \
               f"sid: {self.get_tx_info().short_id}>"



from dataclasses import dataclass
from typing import Union, Optional

from bxcommon.messages.eth.validation.abstract_block_validator import AbstractBlockValidator, \
    BlockValidationResult
from bxcommon.utils.blockchain_utils.eth import eth_common_utils, rlp_utils, eth_common_constants
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging

logger = logging.get_logger(__name__)


@dataclass
class BlockParameters:
    block_hash: Sha256Hash
    block_number: int
    difficulty: int
    mix_hash: Union[bytearray, memoryview]
    proof_of_work_nonce: Union[bytearray, memoryview]


class EthBlockValidator(AbstractBlockValidator):
    def validate_block_header(
        self,
        block_header_bytes: Union[bytearray, memoryview],
        last_confirmed_block_number: Optional[int],
        last_confirmed_block_difficulty: Optional[int]
    ) -> BlockValidationResult:
        if last_confirmed_block_difficulty is None or last_confirmed_block_number is None:
            logger.debug("Skipping block validation because do not have information about last confirmed block")
            return BlockValidationResult(True, None, None)
        block_parameters = self._parse_block_parameters(block_header_bytes)

        valid_block_number = self._validate_block_number(block_parameters, last_confirmed_block_number)
        valid_difficulty = self._validate_block_difficulty(block_parameters, last_confirmed_block_difficulty)
        valid = valid_block_number and valid_difficulty

        if valid:
            logger.debug("Successfully validate block {}", block_parameters.block_hash)
            reason = None
        else:
            logger.info("Validation failed for block {}. Skipping the block.", block_parameters.block_hash)
            reasons = []
            if not valid_block_number:
                reasons.append("invalid number")
            if not valid_difficulty:
                reasons.append("invalid difficulty")
            reason = f"Reason(s) for failure: {', '.join(reasons)}"

        return BlockValidationResult(valid, block_parameters.block_hash, reason)

    def _parse_block_parameters(self, full_block_header_bytes: Union[bytearray, memoryview]) -> BlockParameters:
        _, block_header_len, block_header_start = rlp_utils.consume_length_prefix(full_block_header_bytes, 0)
        block_header_bytes = full_block_header_bytes[block_header_start:block_header_start + block_header_len]
        block_hash = self._get_block_hash(full_block_header_bytes)
        offset = eth_common_constants.FIXED_LENGTH_FIELD_OFFSET
        difficulty, difficulty_length = rlp_utils.decode_int(block_header_bytes, offset)
        offset += difficulty_length
        number, _ = rlp_utils.decode_int(block_header_bytes, offset)
        _gas_limit, gas_limit_length = rlp_utils.decode_int(block_header_bytes, offset)
        offset += gas_limit_length
        _gas_used, gas_used_length = rlp_utils.decode_int(block_header_bytes, offset)
        offset += gas_used_length
        _timestamp, timestamp_length = rlp_utils.decode_int(block_header_bytes, offset)
        offset += timestamp_length
        _, extra_data_length, extra_data_start = rlp_utils.consume_length_prefix(block_header_bytes, offset)
        offset = extra_data_start + extra_data_length
        _, mix_hash_length, mix_hash_start = rlp_utils.consume_length_prefix(block_header_bytes, offset)
        mix_hash = block_header_bytes[mix_hash_start:mix_hash_start + mix_hash_length]
        offset = mix_hash_start + mix_hash_length
        _, nonce_length, nonce_start = rlp_utils.consume_length_prefix(block_header_bytes, offset)
        nonce = block_header_bytes[nonce_start:nonce_start + nonce_length]

        return BlockParameters(block_hash, number, difficulty, mix_hash, nonce)

    def _get_block_hash(self, block_header_bytes: Union[bytearray, memoryview]) -> Sha256Hash:
        raw_hash = eth_common_utils.keccak_hash(memoryview(block_header_bytes))
        return Sha256Hash(raw_hash)

    def _validate_block_number(self, block_parameters: BlockParameters, last_confirmed_block_number: int) -> bool:
        block_number = block_parameters.block_number
        block_hash = block_parameters.block_hash

        valid = block_number - last_confirmed_block_number <= eth_common_constants.MAX_FUTURE_BLOCK_NUMBER
        if not valid:
            logger.debug(
                "Block {} is invalid. Block number {} is way ahead of the last confirmed block {}.",
                block_hash, block_number, last_confirmed_block_number
            )

        return valid

    def _validate_block_difficulty(
        self, block_parameters: BlockParameters, last_confirmed_block_difficulty: int
    ) -> bool:
        block_difficulty = block_parameters.difficulty
        block_hash = block_parameters.block_hash

        max_difficulty_change = int(
            last_confirmed_block_difficulty / 100 * eth_common_constants.MAX_DIFFICULTY_CHANGE_PERCENT
        )

        valid = last_confirmed_block_difficulty - block_difficulty <= max_difficulty_change
        valid = valid or abs(block_difficulty - last_confirmed_block_difficulty) <= 1
        if not valid:
            logger.debug(
                "Block {} is invalid. Block difficulty {} is significantly lower than "
                "the difficulty of last confirmed block {}, with max difficulty change {}.",
                block_hash, block_difficulty, last_confirmed_block_difficulty, max_difficulty_change
            )
        return valid



from typing import Type

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxgateway.messages.btc.addr_btc_message import AddrBtcMessage
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.block_transactions_btc_message import BlockTransactionsBtcMessage
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.messages.btc.data_btc_message import GetBlocksBtcMessage, GetHeadersBtcMessage
from bxgateway.messages.btc.fee_filter_btc_message import FeeFilterBtcMessage
from bxgateway.messages.btc.get_addr_btc_message import GetAddrBtcMessage
from bxgateway.messages.btc.get_block_transactions_btc_message import GetBlockTransactionsBtcMessage
from bxgateway.messages.btc.headers_btc_message import HeadersBtcMessage
from bxgateway.messages.btc.inventory_btc_message import GetDataBtcMessage, InvBtcMessage, NotFoundBtcMessage
from bxgateway.messages.btc.ping_btc_message import PingBtcMessage
from bxgateway.messages.btc.pong_btc_message import PongBtcMessage
from bxgateway.messages.btc.reject_btc_message import RejectBtcMessage
from bxgateway.messages.btc.send_compact_btc_message import SendCompactBtcMessage
from bxgateway.messages.btc.send_headers_btc_message import SendHeadersBtcMessage
from bxgateway.messages.btc.tx_btc_message import TxBtcMessage
from bxgateway.messages.btc.ver_ack_btc_message import VerAckBtcMessage
from bxgateway.messages.btc.version_btc_message import VersionBtcMessage
from bxgateway.messages.btc.xversion_btc_message import XversionBtcMessage


class _BtcMessageFactory(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        BtcMessageType.VERSION: VersionBtcMessage,
        BtcMessageType.VERACK: VerAckBtcMessage,
        BtcMessageType.PING: PingBtcMessage,
        BtcMessageType.PONG: PongBtcMessage,
        BtcMessageType.GET_ADDRESS: GetAddrBtcMessage,
        BtcMessageType.ADDRESS: AddrBtcMessage,
        BtcMessageType.INVENTORY: InvBtcMessage,
        BtcMessageType.GET_DATA: GetDataBtcMessage,
        BtcMessageType.NOT_FOUND: NotFoundBtcMessage,
        BtcMessageType.GET_HEADERS: GetHeadersBtcMessage,
        BtcMessageType.GET_BLOCKS: GetBlocksBtcMessage,
        BtcMessageType.TRANSACTIONS: TxBtcMessage,
        BtcMessageType.BLOCK: BlockBtcMessage,
        BtcMessageType.HEADERS: HeadersBtcMessage,
        BtcMessageType.REJECT: RejectBtcMessage,
        BtcMessageType.SEND_HEADERS: SendHeadersBtcMessage,
        BtcMessageType.COMPACT_BLOCK: CompactBlockBtcMessage,
        BtcMessageType.GET_BLOCK_TRANSACTIONS: GetBlockTransactionsBtcMessage,
        BtcMessageType.BLOCK_TRANSACTIONS: BlockTransactionsBtcMessage,
        BtcMessageType.FEE_FILTER: FeeFilterBtcMessage,
        BtcMessageType.SEND_COMPACT: SendCompactBtcMessage,
        BtcMessageType.XVERSION: XversionBtcMessage
    }

    def __init__(self):
        super(_BtcMessageFactory, self).__init__()
        self.message_type_mapping = self._MESSAGE_TYPE_MAPPING

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return BtcMessage


btc_message_factory = _BtcMessageFactory()



from typing import Optional

from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.ont.data_ont_message import DataOntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class GetHeadersOntMessage(DataOntMessage):
    MESSAGE_TYPE = OntMessageType.GET_HEADERS

    def __init__(self, magic: Optional[int] = None, length: Optional[int] = None,
                 hash_start: Optional[Sha256Hash] = None, hash_stop: Optional[Sha256Hash] = None,
                 buf: Optional[bytearray] = None):
        # pyre-fixme[6]: Expected
        #  `Optional[bxgateway.utils.ont.ont_object_hash.OntObjectHash]` for 3rd param
        #  but got `Optional[Sha256Hash]`.
        super(GetHeadersOntMessage, self).__init__(magic, length, hash_start, hash_stop, self.MESSAGE_TYPE, buf)



import struct
from typing import Optional, List, TYPE_CHECKING

from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

from bxcommon import constants
from bxcommon.constants import BX_HDR_COMMON_OFF
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils import crypto, convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.statistics_event_service import StatisticsEventService

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection


class _BlockStatisticsService(StatisticsEventService):
    BROADCAST_TYPE_OFFSET = BX_HDR_COMMON_OFF + crypto.SHA256_HASH_LEN + \
                            constants.NETWORK_NUM_LEN + constants.NODE_ID_SIZE_IN_BYTES

    def __init__(self) -> None:
        super(_BlockStatisticsService, self).__init__()
        self.name = "BlockInfo"
        self.logger = logging.get_logger(LogRecordType.BlockInfo)
        self.priority_logger = logging.get_logger(LogRecordType.BlockPropagationInfo)
        self.log_detailed_block_stats = False

    def set_node(self, node) -> None:
        super(_BlockStatisticsService, self).set_node(node)
        assert node.opts is not None
        self.log_detailed_block_stats = node.opts.log_detailed_block_stats

    def add_block_event(
        self,
        block_msg,
        block_event_settings,
        network_num,
        start_date_time=None,
        end_date_time=None,
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ) -> None:
        if not self._should_log_stat_event(block_event_settings):
            return

        if isinstance(block_msg, BroadcastMessage):
            block_hash = block_msg.block_hash().binary
            broadcast_type = block_msg.broadcast_type().value
        elif isinstance(block_msg, memoryview):
            block_hash = block_msg[BX_HDR_COMMON_OFF:BX_HDR_COMMON_OFF + crypto.SHA256_HASH_LEN].tobytes()
            broadcast_type = struct.unpack_from(
                "<4s", block_msg[self.BROADCAST_TYPE_OFFSET:
                                 self.BROADCAST_TYPE_OFFSET + constants.BROADCAST_TYPE_LEN].tobytes()
            )[0].decode(constants.DEFAULT_TEXT_ENCODING)
        else:
            block_hash = block_msg[BX_HDR_COMMON_OFF:BX_HDR_COMMON_OFF + crypto.SHA256_HASH_LEN]
            broadcast_type = struct.unpack_from(
                "<4s", block_msg[self.BROADCAST_TYPE_OFFSET:self.BROADCAST_TYPE_OFFSET + constants.BROADCAST_TYPE_LEN]
            )[0].decode(constants.DEFAULT_TEXT_ENCODING)

        self.log_event(
            block_event_settings, convert.bytes_to_hex(block_hash), start_date_time, end_date_time,
            network_num=network_num, broadcast_type=broadcast_type, peers=peers, **kwargs
        )

    def add_block_event_by_block_hash(
        self,
        block_hash,
        block_event_settings,
        network_num,
        broadcast_type=BroadcastMessageType.BLOCK,
        start_date_time=None,
        end_date_time=None,
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ) -> None:
        if not self._should_log_stat_event(block_event_settings):
            return

        if isinstance(block_hash, Sha256Hash):
            block_hash_str = block_hash.binary
        elif isinstance(block_hash, memoryview):
            block_hash_str = block_hash.tobytes()
        else:
            block_hash_str = block_hash

        self.log_event(
            block_event_settings,
            convert.bytes_to_hex(block_hash_str),
            start_date_time,
            end_date_time,
            network_num=network_num,
            broadcast_type=broadcast_type.value,
            peers=peers, **kwargs)

    def _should_log_stat_event(self, event_type_settings) -> bool:
        return self.log_detailed_block_stats or not event_type_settings.detailed_stat_event


block_stats = _BlockStatisticsService()



import struct
from typing import Optional, List, Union

from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ont_messages_util import ont_varint_to_int, pack_int_to_ont_varint, get_txid


class BlockOntMessage(OntMessage, AbstractBlockMessage):
    MESSAGE_TYPE = OntMessageType.BLOCK

    def __init__(self, magic: Optional[int] = None, version: Optional[int] = None,
                 prev_block: Optional[OntObjectHash] = None, txns_root: Optional[OntObjectHash] = None,
                 block_root: Optional[OntObjectHash] = None, timestamp: Optional[int] = None,
                 height: Optional[int] = None, consensus_data: Optional[int] = None,
                 consensus_payload: Optional[bytes] = None, next_bookkeeper: Optional[bytes] = None,
                 bookkeepers: Optional[list] = None, sig_data: Optional[list] = None, txns: Optional[list] = None,
                 merkle_root: Optional[OntObjectHash] = None, buf: Optional[Union[bytearray, memoryview]] = None):
        if buf is None:
            # pyre-fixme[16]: `Optional` has no attribute `__iter__`.
            total_tx_size = sum(len(tx) for tx in txns)
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + len(consensus_payload) +
                            # pyre-fixme[6]: Expected `Sized` for 1st param but got
                            #  `Optional[List[typing.Any]]`.
                            ont_constants.ONT_BOOKKEEPER_AND_VARINT_LEN * len(bookkeepers) +
                            # pyre-fixme[6]: Expected `Sized` for 1st param but got
                            #  `Optional[List[typing.Any]]`.
                            # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
                            len(sig_data) * (len(sig_data[0]) + ont_constants.ONT_VARINT_MAX_LEN) + total_tx_size + 188
                            + ont_constants.ONT_VARINT_MAX_LEN * 3)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<I", buf, off, version)
            off += ont_constants.ONT_INT_LEN
            # pyre-fixme[16]: `Optional` has no attribute `get_little_endian`.
            buf[off:off + ont_constants.ONT_HASH_LEN] = prev_block.get_little_endian()
            off += ont_constants.ONT_HASH_LEN
            buf[off:off + ont_constants.ONT_HASH_LEN] = txns_root.get_little_endian()
            off += ont_constants.ONT_HASH_LEN
            buf[off:off + ont_constants.ONT_HASH_LEN] = block_root.get_little_endian()
            off += ont_constants.ONT_HASH_LEN
            struct.pack_into("<IIQ", buf, off, timestamp, height, consensus_data)
            off += ont_constants.ONT_BLOCK_TIME_HEIGHT_CONS_DATA_LEN

            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            off += pack_int_to_ont_varint(len(consensus_payload), buf, off)
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for 2nd
            #  param but got `Optional[bytes]`.
            buf[off:off + len(consensus_payload)] = consensus_payload
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            off += len(consensus_payload)
            struct.pack_into("<20s", buf, off, next_bookkeeper)
            off += ont_constants.ONT_BLOCK_NEXT_BOOKKEEPER_LEN

            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[typing.Any]]`.
            off += pack_int_to_ont_varint(len(bookkeepers), buf, off)
            for bk in bookkeepers:
                off += pack_int_to_ont_varint(len(bk), buf, off)
                buf[off:off + ont_constants.ONT_BOOKKEEPER_LEN] = bk
                off += ont_constants.ONT_BOOKKEEPER_LEN

            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[typing.Any]]`.
            off += pack_int_to_ont_varint(len(sig_data), buf, off)
            for sig in sig_data:
                off += pack_int_to_ont_varint(len(sig), buf, off)
                buf[off:off + len(sig)] = sig
                off += len(sig)

            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[typing.Any]]`.
            struct.pack_into("<L", buf, off, len(txns))
            off += ont_constants.ONT_INT_LEN
            for tx in txns:
                buf[off:off + len(tx)] = tx
                off += len(tx)

            buf[off:off + ont_constants.ONT_HASH_LEN] = merkle_root.get_little_endian()
            off += ont_constants.ONT_HASH_LEN

            super(BlockOntMessage, self).__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            assert not isinstance(buf, str)
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = None
        self._prev_block: Optional[OntObjectHash] = None
        self._txns_root = self._block_root = self._height = self._consensus_data = self._consensus_payload = None
        self._next_bookkeeper = self._txns = self._merkle_root = self._consensus_payload_length = None
        self._header = self._header_offset = self._header_with_program = self._header_with_program_offset = None
        self._hash_val = self._bookkeepers_length = self._sig_data_length = None
        self._txn_count = self._tx_offset = self._txn_header = None
        self._merkle_root_memoryview = None
        self._timestamp = 0
        self._parsed = False

    def __repr__(self):
        return "BlockOntMessage<block_hash: {}, length: {}>".format(self.block_hash(), len(self.rawbytes()))

    def parse_message(self) -> int:
        if not self._parsed:
            self._parsed = True
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._version, = struct.unpack_from("<I", self.buf, off)
            off += ont_constants.ONT_INT_LEN
            self._prev_block = OntObjectHash(self.buf, off, ont_constants.ONT_HASH_LEN)
            off += ont_constants.ONT_HASH_LEN
            self._txns_root = OntObjectHash(self.buf, off, ont_constants.ONT_HASH_LEN)
            off += ont_constants.ONT_HASH_LEN
            self._block_root = OntObjectHash(self.buf, off, ont_constants.ONT_HASH_LEN)
            off += ont_constants.ONT_HASH_LEN
            self._timestamp, self._height, self._consensus_data = struct.unpack_from("<IIQ", self.buf, off)
            off += ont_constants.ONT_BLOCK_TIME_HEIGHT_CONS_DATA_LEN
            self._consensus_payload_length, size = ont_varint_to_int(self.buf, off)
            # pyre-fixme[6]: Expected `int` for 1st param but got `None`.
            off += self._consensus_payload_length + size
            self._next_bookkeeper, = struct.unpack_from("<20s", self.buf, off)
            off += ont_constants.ONT_BLOCK_NEXT_BOOKKEEPER_LEN

            self._header_offset = off
            self._header = self._memoryview[0:self._header_offset]

            self._bookkeepers_length, size = ont_varint_to_int(self.buf, off)
            off += size
            # pyre-fixme[6]: Expected `int` for 1st param but got `None`.
            for _ in range(self._bookkeepers_length):
                _, size = ont_varint_to_int(self.buf, off)
                off += size + ont_constants.ONT_BOOKKEEPER_LEN

            self._sig_data_length, size = ont_varint_to_int(self.buf, off)
            off += size
            # pyre-fixme[6]: Expected `int` for 1st param but got `None`.
            for _ in range(self._sig_data_length):
                sig_length, size = ont_varint_to_int(self.buf, off)
                off += size + sig_length

            self._header_with_program_offset = off
            self._header_with_program = self._memoryview[ont_constants.ONT_HDR_COMMON_OFF:
                                                         self._header_with_program_offset]

            self._txn_count, = struct.unpack_from("<L", self.buf, off)
            off += ont_constants.ONT_INT_LEN
            self._tx_offset = off
            self._txn_header = self._memoryview[0:self._tx_offset]

        version = self._version
        assert isinstance(version, int)
        return version

    def prev_block_hash(self) -> OntObjectHash:
        if not self._parsed:
            self.parse_message()

        prev_block = self._prev_block
        assert prev_block is not None
        return prev_block

    def timestamp(self) -> int:
        """
        :return: seconds since epoch
        """
        assert self._timestamp is not None
        if not self._parsed:
            self.parse_message()
        return self._timestamp

    def txn_count(self) -> int:
        if not self._parsed:
            self.parse_message()
        txn_count = self._txn_count
        assert isinstance(txn_count, int)
        return txn_count

    def txn_offset(self) -> int:
        if not self._parsed:
            self.parse_message()
        tx_offset = self._tx_offset
        assert isinstance(tx_offset, int)
        return tx_offset

    def txn_header(self) -> memoryview:
        if not self._parsed:
            self.parse_message()
            self._txn_header = self._memoryview[0:self._tx_offset]
        txn_header = self._txn_header
        assert isinstance(txn_header, memoryview)
        return txn_header

    def txns(self) -> List[memoryview]:
        if self._txns is None:
            if self._tx_offset is None:
                self.parse_message()
            assert self._tx_offset is not None
            start = self._tx_offset
            self._txns = []
            for _ in range(self.txn_count()):
                _, off = get_txid(self._memoryview[start:])
                sig_length, size = ont_varint_to_int(self._memoryview[start:], off)
                off += size
                for _ in range(sig_length):
                    invoke_length, size = ont_varint_to_int(self._memoryview[start:], off)
                    off += size + invoke_length
                    verify_length, size = ont_varint_to_int(self._memoryview[start:], off)
                    off += size + verify_length
                # pyre-fixme[16]: Optional type has no attribute `append`.
                self._txns.append(self._memoryview[start:start + off])
                start += off
            self._merkle_root = OntObjectHash(self._memoryview[start:], 0, ont_constants.ONT_HASH_LEN)
            self._merkle_root_memoryview = self._memoryview[start:start + ont_constants.ONT_HASH_LEN]
        txns = self._txns
        assert isinstance(txns, list)
        return txns

    def block_header_offset(self) -> int:
        if not self._parsed:
            self.parse_message()
        header_offset = self._header_offset
        assert isinstance(header_offset, int)
        return header_offset

    def block_hash(self) -> OntObjectHash:
        if self._hash_val is None:
            if self._header_offset is None:
                self.parse_message()
            assert self._header_offset is not None
            header = self._memoryview[ont_constants.ONT_HDR_COMMON_OFF:self._header_offset]
            raw_hash = crypto.double_sha256(header)
            self._hash_val = OntObjectHash(buf=raw_hash, length=ont_constants.ONT_HASH_LEN)
        hash_val = self._hash_val
        assert isinstance(hash_val, OntObjectHash)
        return hash_val

    def header(self) -> memoryview:
        if self._header_with_program is None:
            self.parse_message()
        header_with_program = self._header_with_program
        assert isinstance(header_with_program, memoryview)
        return header_with_program

    def height(self) -> int:
        if self._height is None:
            self.parse_message()
        height = self._height
        assert isinstance(height, int)
        return height

    def merkle_root(self) -> memoryview:
        if self._merkle_root_memoryview is None:
            self.txns()
        merkle_root_memoryview = self._merkle_root_memoryview
        assert isinstance(merkle_root_memoryview, memoryview)
        return merkle_root_memoryview

    def version(self) -> int:
        if self._version is None:
            off = ont_constants.ONT_HDR_COMMON_OFF
            self._version, = struct.unpack_from("<I", self.buf, off)
        version = self._version
        assert isinstance(version, int)
        return version



class GatewayMessageType:
    HELLO = b"gw_hello"
    BLOCK_RECEIVED = b"blockrecv"
    BLOCK_PROPAGATION_REQUEST = b"blockprop"

    # Sync messages types are currently unused. See `blockchain_sync_service.py`.
    SYNC_REQUEST = b"syncreq"
    SYNC_RESPONSE = b"syncres"

    REQUEST_TX_STREAM = b"rqtx"
    CONFIRMED_TX = b"cnfrmtx"
    CONFIRMED_BLOCK = b"cnfrmbk"



import asyncio
import base64
import os
import ssl
from collections import defaultdict
from ssl import Purpose
from abc import abstractmethod
from asyncio import Future
from typing import Callable, Awaitable, Optional, TYPE_CHECKING, TypeVar, Generic, List, Dict, cast
from aiohttp import web
from aiohttp.abc import StreamResponse
from aiohttp.web import Application, Request, Response, AppRunner, TCPSite, WebSocketResponse
from aiohttp.web_exceptions import HTTPClientError, HTTPBadRequest, HTTPInternalServerError
from aiohttp.web_exceptions import HTTPUnauthorized

from bxcommon import constants
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.abstract_ws_rpc_handler import AbstractWsRpcHandler
from bxcommon.rpc.https.http_rpc_handler import HttpRpcHandler
from bxcommon.rpc.https.request_formatter import RequestFormatter
from bxcommon.rpc.https.response_formatter import ResponseFormatter
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_constants import ContentType
from bxcommon.rpc.ws.ws_connection import WsConnection

from bxcommon.rpc.rpc_errors import RpcError, RpcParseError, RpcMethodNotFound, \
    RpcInvalidParams, RpcAccountIdError
from bxutils.encoding import json_encoder

from bxutils import logging

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
Node = TypeVar("Node", bound="AbstractNode")


@web.middleware
async def request_middleware(
    request: Request, handler: Callable[[Request], Awaitable[StreamResponse]]
) -> StreamResponse:
    request_formatter = RequestFormatter(request)
    logger.trace("Handling RPC request: {}.", request_formatter)
    response = await handler(request)
    if isinstance(response, Response):
        logger.trace(
            "Finished handling request: {}, returning response: {}.",
            request_formatter,
            ResponseFormatter(response),
        )
    else:
        logger.trace(
            "Finished handling web scoket: {}.",
            request_formatter,
        )

    return response


def format_http_error(
    client_error: HTTPClientError, content_type: ContentType
) -> HTTPClientError:
    err_msg = client_error.text
    code = client_error.status_code
    response_json = {
        "result": None,
        "error": err_msg,
        "code": code,
        "message": err_msg,
    }
    client_error.content_type = content_type.value
    client_error.text = json_encoder.to_json(response_json)
    return client_error


def format_rpc_error(
    rpc_error: RpcError, status_code: int, content_type: ContentType
) -> Response:
    if content_type == ContentType.PLAIN:
        return web.json_response(
            JsonRpcResponse(rpc_error.id, error=rpc_error).to_jsons(),
            status=status_code,
            content_type=ContentType.PLAIN.value
        )
    else:
        return web.json_response(
            text=JsonRpcResponse(rpc_error.id, error=rpc_error).to_jsons(),
            status=status_code,
        )


class AbstractHttpRpcServer(Generic[Node]):
    RUN_SLEEP_INTERVAL_S: int = 5

    node: Node
    _app: Application
    _runner: AppRunner
    _site: Optional[TCPSite]
    _handler: HttpRpcHandler[Node]
    _stop_requested: bool
    _stop_waiter: Future
    _started: bool
    _encoded_auth: Optional[str]
    _ws_connections: Dict[str, List[WsConnection]] = defaultdict(list)

    def __init__(
        self,
        node: Node,
    ) -> None:
        self.node = node
        self._app = Application(middlewares=[request_middleware])
        self._app.add_routes(
            [web.post("/", self.handle_request), web.get("/", self.handle_get_request)]
        )
        self._app.add_routes(
            [web.post("/ws", self.handle_ws_request),
             web.get("/ws", self.handle_ws_request)]
        )
        self._runner = AppRunner(self._app)
        self._site = None
        self._handler = self.request_handler()
        self._stop_requested = False
        self._stop_waiter = asyncio.get_event_loop().create_future()
        self._started = False
        self._encoded_auth = None
        self._ws_connections: Dict[str, List[WsConnection]] = defaultdict(list)

        rpc_user = self.node.opts.rpc_user
        if rpc_user:
            rpc_password = self.node.opts.rpc_password
            self._encoded_auth = base64.b64encode(
                f"{rpc_user}:{rpc_password}".encode("utf-8")
            ).decode("utf-8")

    @abstractmethod
    def authenticate_request(self, request: Request) -> None:
        pass

    @abstractmethod
    def request_handler(self) -> HttpRpcHandler:
        pass

    @abstractmethod
    def request_ws_handler(self) -> Optional[AbstractWsRpcHandler]:
        pass

    def status(self) -> bool:
        return self._started

    async def run(self) -> None:
        try:
            await self._start()
            while not self._stop_requested:
                await asyncio.sleep(self.RUN_SLEEP_INTERVAL_S)
        finally:
            self._stop_waiter.set_result(True)

    async def start(self) -> None:
        if self._started:
            return
        try:
            await self._start()
        finally:
            self._stop_waiter.set_result(True)

    async def stop(self) -> None:
        self._stop_requested = True
        if self._started:
            await asyncio.gather(
                *(connection.close() for ip, connections in self._ws_connections.items() for connection in connections)
            )
        await self._stop_waiter
        await self._runner.cleanup()
        self._started = False

    async def handle_request(self, request: Request) -> Response:
        try:
            self._handler.parse_content_type(request)
            self.authenticate_request(request)
            return await self._handler.handle_request(request)
        except HTTPClientError as e:
            return format_http_error(e, self._handler.content_type)
        except RpcAccountIdError as e:
            return format_rpc_error(e, HTTPUnauthorized.status_code, self._handler.content_type)
        except (RpcParseError, RpcMethodNotFound, RpcInvalidParams) as e:
            return format_rpc_error(e, HTTPBadRequest.status_code, self._handler.content_type)
        except RpcError as e:
            return format_rpc_error(e, HTTPInternalServerError.status_code, self._handler.content_type)

    async def handle_get_request(self, request: Request) -> Response:
        try:
            self.authenticate_request(request)
        except HTTPUnauthorized as e:
            return format_http_error(e, self._handler.content_type)
        except RpcAccountIdError as e:
            return format_rpc_error(
                e,
                HTTPUnauthorized.status_code,
                self._handler.content_type
            )
        else:
            response_dict = {
                "result": {
                    "required_request_type": "POST",
                    "required_headers": [
                        {
                            rpc_constants.CONTENT_TYPE_HEADER_KEY: ContentType.PLAIN.value
                        }
                    ],
                    "payload_structures": await self._handler.help(),
                }
            }
            json_response = JsonRpcResponse.from_json(response_dict)
            return web.json_response(json_response.to_jsons(), dumps=json_encoder.to_json)

    async def handle_ws_request(self, request: Request) -> WebSocketResponse:
        try:
            self.authenticate_request(request)
        except RpcError as e:
            websocket_response = web.WebSocketResponse()
            await websocket_response.prepare(request)
            error_message = e.data
            assert error_message is not None
            # WebSockets close event code 1008: Policy Violation.
            await websocket_response.close(
                code=1008, message=bytes(error_message.encode(constants.DEFAULT_TEXT_ENCODING))
            )
            return websocket_response
        else:
            ws_response = web.WebSocketResponse()
            websocket_handler = self.request_ws_handler()
            assert websocket_handler is not None
            path = cast(str, request.path)
            ws_connection = WsConnection(
                ws_response,
                path,
                websocket_handler
            )
            # casting for pyre check tests
            remote_ip = cast(str, request.remote)
            if remote_ip is None:
                remote_ip = rpc_constants.DEFAULT_RPC_HOST
            # this coroutine may close by the http server. use try/finally
            # to force cleanup
            try:
                self._ws_connections[remote_ip].append(ws_connection)
                websocket_response = await ws_connection.handle(request)
            finally:
                await ws_connection.close()
                self._ws_connections[remote_ip].remove(ws_connection)
                logger.debug("websocket connection for {} closed and removed", remote_ip)

            return websocket_response

    async def _start(self) -> None:
        self._started = True
        await self._runner.setup()
        opts = self.node.opts

        # TODO: add ssl certificate
        ssl_context = None
        if opts.rpc_use_ssl:
            ssl_context = ssl.create_default_context(
                Purpose.CLIENT_AUTH,
                cafile=os.path.join(opts.rpc_ssl_base_url, "ca_bundle.pem")
            )
            ssl_context.load_cert_chain(
                certfile=os.path.join(opts.rpc_ssl_base_url, "cert.pem"),
                keyfile=os.path.join(opts.rpc_ssl_base_url, "key.pem")
            )
            ssl_context.check_hostname = False
        logger.info("Starting listening on RPC {}:{} SSL:{}",
                    opts.rpc_host, opts.rpc_port, opts.rpc_use_ssl)
        site = TCPSite(self._runner, opts.rpc_host, opts.rpc_port, ssl_context=ssl_context)
        self._site = site
        await site.start()



from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.messages.bloxroute.bloxroute_message_factory import _BloxrouteMessageFactory
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxgateway.messages.gateway.v1.gateway_hello_message_v1 import GatewayHelloMessageV1
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class _GatewayMessageFactoryV1(_BloxrouteMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        GatewayMessageType.HELLO: GatewayHelloMessageV1,
        BloxrouteMessageType.ACK: AckMessage,
    }


gateway_message_factory_v1 = _GatewayMessageFactoryV1()



from typing import Optional, List, Union, Any, Dict, Tuple

from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.rpc.rpc_request_type import RpcRequestType


class WsClient(AbstractWsProvider):
    async def call_bx(
        self,
        method: RpcRequestType,
        params: Union[List[Any], Dict[Any, Any], None],
        request_id: Optional[str] = None
    ) -> JsonRpcResponse:
        if request_id is None:
            request_id = str(self.current_request_id)
            self.current_request_id += 1

        return await self.call(
            BxJsonRpcRequest(request_id, method, params)
        )

    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        if options is None:
            options = {}
        response = await self.call_bx(
            RpcRequestType.SUBSCRIBE, [channel, options]
        )
        subscription_id = response.result
        assert isinstance(subscription_id, str)
        self.subscription_manager.register_subscription(subscription_id)
        return subscription_id

    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        response = await self.call_bx(RpcRequestType.UNSUBSCRIBE, [subscription_id])
        if response.result is not None:
            self.subscription_manager.unregister_subscription(subscription_id)
            return True, None
        else:
            return False, response.error



import time
import typing
from abc import ABC
from typing import TYPE_CHECKING, TypeVar

from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.utils import memory_utils
from bxcommon.utils.stats import stats_format, hooks
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxgateway import gateway_constants
from bxutils import logging
from bxgateway import log_messages

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

GatewayNode = TypeVar("GatewayNode", bound="AbstractGatewayNode")

logger = logging.get_logger(__name__)


class AbstractGatewayBlockchainConnection(AbstractConnection[GatewayNode], ABC):
    CONNECTION_TYPE = ConnectionType.BLOCKCHAIN_NODE

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: GatewayNode):
        super(AbstractGatewayBlockchainConnection, self).__init__(sock, node)

        if node.opts.tune_send_buffer_size:
            transport = sock.transport
            assert transport is not None

            # Requires OS tuning (for Linux here; MacOS and windows require different commands):
            # echo 'net.core.wmem_max=16777216' >> /etc/sysctl.conf && sysctl -p
            # cat /proc/sys/net/core/wmem_max # to check

            # pyre-fixme[16]: `Transport` has no attribute `get_write_buffer_limits`.
            _, previous_buffer_size = transport.get_write_buffer_limits()
            try:
                transport.set_write_buffer_limits(high=gateway_constants.BLOCKCHAIN_SOCKET_SEND_BUFFER_SIZE)
                self.log_debug(
                    "Setting socket send buffer size for blockchain connection to {}",
                    transport.get_write_buffer_limits()
                )
            except Exception as e:
                self.log_warning(log_messages.SOCKET_BUFFER_SEND_FAIL, e)

            # Linux systems generally set the value to 2x what was passed in, so just make sure
            # the socket buffer is at least the size set
            _, set_buffer_size = transport.get_write_buffer_limits()
            if set_buffer_size < gateway_constants.BLOCKCHAIN_SOCKET_SEND_BUFFER_SIZE:
                self.log_warning(log_messages.SET_SOCKET_BUFFER_SIZE, set_buffer_size, previous_buffer_size)
                # transport.set_write_buffer_limits(high=previous_buffer_size)

        self.connection_protocol = None
        self.is_server = False

        self.can_send_pings = True
        self.pong_timeout_enabled = True

    def __hash__(self):
        return hash((self.endpoint.ip_address, self.endpoint.port))

    def __eq__(self, other) -> bool:
        if not isinstance(other, AbstractGatewayBlockchainConnection):
            return False
        return self.endpoint.ip_address == other.endpoint.ip_address and self.endpoint.port == other.endpoint.port

    def log_connection_mem_stats(self) -> None:
        """
        logs the connection's memory stats
        """
        super(AbstractGatewayBlockchainConnection, self).log_connection_mem_stats()
        t0 = time.time()
        class_name = self.__class__.__name__
        if self.node.message_converter is not None:
            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self.node.message_converter,
                "message_converter",
                memory_utils.ObjectSize(
                    "message_converter", memory_utils.get_special_size(self.node.message_converter).size,
                    is_actual_size=True
                ),
                object_item_count=1,
                object_type=memory_utils.ObjectType.META,
                size_type=memory_utils.SizeType.SPECIAL
            )
        t_m_c_diff = round(time.time() - t0, 3)
        logger.debug("Recording {} message_converter MemoryStats took: {} seconds", class_name, t_m_c_diff)

    def get_connection_state_details(self):
        """
        Returns details of the current connection state. Used to submit details to SDN on disconnect.
        :return:
        """
        return f"Connection state: {self.state}"

    def dispose(self):

        if self.CONNECTION_TYPE == ConnectionType.BLOCKCHAIN_NODE:
            self.node.on_blockchain_connection_destroyed(self)
        elif self.CONNECTION_TYPE == ConnectionType.REMOTE_BLOCKCHAIN_NODE:
            if self.node.remote_node_conn is None or self.node.remote_node_conn == self:
                self.node.on_remote_blockchain_connection_destroyed(self)
            else:
                logger.warning(log_messages.CLOSE_CONNECTION_ATTEMPT,
                               self.peer_desc, self.node.remote_node_conn.peer_desc)
        super().dispose()



from typing import List

from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_SHA_HASH_LEN, BTC_VARINT_MIN_SIZE
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint


class GetBlockTransactionsBtcMessage(BtcMessage):
    """
    Message used to request from peer transactions in a Compact Block that were not found in memory pool
    """

    MESSAGE_TYPE = BtcMessageType.GET_BLOCK_TRANSACTIONS

    # pyre-fixme[9]: magic has type `int`; used as `None`.
    # pyre-fixme[9]: block_hash has type `BtcObjectHash`; used as `None`.
    # pyre-fixme[9]: indices has type `List[int]`; used as `None`.
    def __init__(self, magic: int = None, block_hash: BtcObjectHash = None, indices: List[int] = None,
                 # pyre-fixme[9]: buf has type `memoryview`; used as `None`.
                 buf: memoryview = None):
        if buf is None:
            # pyre-fixme[9]: buf has type `memoryview`; used as `bytearray`.
            buf = bytearray(
                BTC_HDR_COMMON_OFF + BTC_SHA_HASH_LEN + BTC_VARINT_MIN_SIZE + BTC_VARINT_MIN_SIZE * len(indices))
            off = BTC_HDR_COMMON_OFF

            buf[off:off + BTC_SHA_HASH_LEN] = block_hash.get_big_endian()
            off += BTC_SHA_HASH_LEN

            off += pack_int_to_btc_varint(len(indices), buf, off)

            last_real_index = -1
            for real_index in indices:
                diff = real_index - last_real_index - 1
                last_real_index = real_index
                off += pack_int_to_btc_varint(diff, buf, off)

            self.buf = buf
            super(GetBlockTransactionsBtcMessage, self).__init__(magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF,
                                                                 buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._block_hash = None
        # pyre-fixme[8]: Attribute has type `List[int]`; used as `None`.
        self._indices: List[int] = None

    def block_hash(self) -> BtcObjectHash:
        if self._block_hash is None:
            self._block_hash = BtcObjectHash(buf=self.buf, offset=BTC_HDR_COMMON_OFF, length=BTC_SHA_HASH_LEN)

        # pyre-fixme[7]: Expected `BtcObjectHash` but got `None`.
        return self._block_hash

    def indices(self) -> List[int]:
        if self._indices is None:
            off = BTC_HDR_COMMON_OFF + 32
            num_indices, size = btc_varint_to_int(self.buf, off)
            off += size
            self._indices = []

            index = -1
            for _ in range(num_indices):
                diff_index, size = btc_varint_to_int(self.buf, off)
                index += diff_index + 1
                self._indices.append(index)
                off += size

        return self._indices



from typing import Dict

from bxcommon.rpc.provider.abstract_provider import SubscriptionNotification
from bxcommon.utils.async_live_queue import AsyncLiveQueue


class SubscriptionManager:
    queue_limit: int
    all_subscription_notifications: AsyncLiveQueue[SubscriptionNotification]
    subscription_notifications_by_id: Dict[str, AsyncLiveQueue[SubscriptionNotification]]

    def __init__(self, queue_limit: int):
        self.queue_limit = queue_limit
        self.all_subscription_notifications = AsyncLiveQueue(self.queue_limit)
        self.subscription_notifications_by_id = {}

    def register_subscription(self, subscription_id: str) -> None:
        self.subscription_notifications_by_id[subscription_id] = AsyncLiveQueue(self.queue_limit)

    def unregister_subscription(self, subscription_id: str) -> None:
        del self.subscription_notifications_by_id[subscription_id]

    async def receive_message(self, subscription_notification: SubscriptionNotification) -> None:
        """
        Due to asynchronous scheduling it's possible for a message to be received
        before the subscription is registered. These messages will be dropped.
        """
        await self.all_subscription_notifications.put(subscription_notification)
        if subscription_notification.subscription_id in self.subscription_notifications_by_id:
            await self.subscription_notifications_by_id[
                subscription_notification.subscription_id
            ].put(subscription_notification)

    async def get_next_subscription_notification(self) -> SubscriptionNotification:
        """
        NOTE: this does not consume notifications from subscription id specific queues.
        :return:
        """
        return await self.all_subscription_notifications.get()

    async def get_next_subscription_notification_for_id(
        self, subscription_id: str
    ) -> SubscriptionNotification:
        return await self.subscription_notifications_by_id[subscription_id].get()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Union, Dict, List, cast, Iterator, Tuple

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.network.ip_endpoint import IpEndpoint
from bxgateway.utils.logging.status import summary
from bxgateway.utils.logging.status.blockchain_connection import BlockchainConnection
from bxgateway.utils.logging.status.connection_state import ConnectionState
from bxgateway.utils.logging.status.gateway_status import GatewayStatus
from bxgateway.utils.logging.status.relay_connection import RelayConnection
from bxgateway.utils.logging.status.summary import Summary


@dataclass
class Network:
    relays: List[RelayConnection]
    blockchain_nodes: List[BlockchainConnection]
    remote_blockchain_nodes: List[BlockchainConnection]

    def iter_network_type_pairs(
        self
    ) -> Iterator[Tuple[ConnectionType, Union[List[RelayConnection], List[BlockchainConnection]]]]:
        networks = {
            ConnectionType.RELAY_ALL: self.relays,
            ConnectionType.BLOCKCHAIN_NODE: self.blockchain_nodes,
            ConnectionType.REMOTE_BLOCKCHAIN_NODE: self.remote_blockchain_nodes
        }
        for network_type, network in networks.items():
            yield network_type, network

    def get_summary(self, ip_address: str, continent: str, country: str, update_required: bool,
                    account_id: Optional[str], quota_level: Optional[int]) -> Summary:
        relay_connections_state = _connections_states_info(self.relays)
        blockchain_node_connections_state = _blockchain_connections_state_info(self.blockchain_nodes)
        remote_blockchain_node_connections_state = _connections_states_info(self.remote_blockchain_nodes)
        gateway_status = self._get_gateway_status()

        return Summary(gateway_status, summary.gateway_status_get_account_info(account_id),
                       relay_connections_state,
                       blockchain_node_connections_state, remote_blockchain_node_connections_state,
                       ip_address, continent, country, update_required,
                       summary.gateway_status_get_quota_level(quota_level))

    def remove_connection(self, conn: Union[RelayConnection, BlockchainConnection], conn_type: ConnectionType) -> None:
        if conn_type == ConnectionType.RELAY_ALL:
            self.relays.remove(cast(RelayConnection, conn))
        elif conn_type == ConnectionType.BLOCKCHAIN_NODE:
            self.blockchain_nodes.remove(cast(BlockchainConnection, conn))
        elif conn_type == ConnectionType.REMOTE_BLOCKCHAIN_NODE:
            self.remote_blockchain_nodes.remove(cast(BlockchainConnection, conn))

    def add_connection(self, conn: ConnectionType, desc: str, file_no: Optional[str] = None,
                       peer_id: Optional[str] = None) -> None:
        ip_addr = desc.split()[0]
        port = desc.split()[1]
        current_time = _get_current_time()

        if conn == ConnectionType.RELAY_ALL:
            relay_connection = RelayConnection(ip_addr, port, file_no, peer_id, current_time)
            if relay_connection not in self.relays:
                self.relays.append(relay_connection)
        elif conn == ConnectionType.BLOCKCHAIN_NODE:
            assert ip_addr is not None
            assert port is not None
            blockchain_connection = BlockchainConnection(ip_addr, port, file_no, current_time)
            if blockchain_connection not in self.blockchain_nodes:
                self.blockchain_nodes.append(blockchain_connection)
        elif conn == ConnectionType.REMOTE_BLOCKCHAIN_NODE:
            assert ip_addr is not None
            assert port is not None
            blockchain_connection = BlockchainConnection(ip_addr, port, file_no, current_time)
            if blockchain_connection not in self.remote_blockchain_nodes:
                self.remote_blockchain_nodes.append(BlockchainConnection(ip_addr, port, file_no, current_time))

    def _get_gateway_status(self) -> GatewayStatus:
        return GatewayStatus.ONLINE if _check_connections_established(
            self.relays) and _check_connections_established(
            self.blockchain_nodes) and _check_connections_established(
            self.remote_blockchain_nodes) else GatewayStatus.WITH_ERRORS


def _get_current_time() -> str:
    return "UTC " + str(datetime.utcnow())


def _check_connections_established(connections: Union[List[RelayConnection], List[BlockchainConnection]]) -> bool:
    return len(connections) > 0 and all([conn.get_connection_state() == ConnectionState.ESTABLISHED for conn in connections])


def _connections_states_info(connections: Union[List[RelayConnection], List[BlockchainConnection]]) -> ConnectionState:
    return ConnectionState.ESTABLISHED if _check_connections_established(connections) else ConnectionState.DISCONNECTED


def _blockchain_connections_state_info(connections: List[BlockchainConnection]) -> Dict[str, ConnectionState]:
    connection_states = {}
    for conn in connections:
        connection_states[str(IpEndpoint(conn.ip_address, int(conn.port)))] = \
            ConnectionState.ESTABLISHED if _check_connections_established([conn]) else ConnectionState.DISCONNECTED
    return connection_states



from abc import ABCMeta, abstractmethod

from bxcommon.messages.abstract_internal_message import AbstractInternalMessage


class AbstractMessageConverter:
    __metaclass__ = ABCMeta

    @abstractmethod
    def convert_to_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        """
        Converts message of current protocol version to older version (version of converter)

        :param msg: message in current protocol version
        :return: converted message
        """

    @abstractmethod
    def convert_from_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        """
        Converts message of older version (version of converter) to current protocol version

        :param msg: message in older protocol version
        :return: converted message
        """

    @abstractmethod
    def convert_first_bytes_to_older_version(self, first_msg_bytes):
        """
        Converts first bytes of message of current protocol version to older version (version of converter)

        :param first_msg_bytes: first bytes of message in current protocol version
        :return: converted message bytes
        """

    @abstractmethod
    def convert_first_bytes_from_older_version(self, first_msg_bytes):
        """
        Converts first message bytes of older version (version of converter) to current protocol version

        :param first_msg_bytes: first message bytes in older protocol version
        :return: converted message bytes
        """

    @abstractmethod
    def convert_last_bytes_to_older_version(self, last_msg_bytes):
        """
        Converts last bytes of message of current protocol version to older version (version of converter)

        :param last_msg_bytes: first bytes of message in current protocol version
        :return: converted message bytes
        """

    @abstractmethod
    def convert_last_bytes_from_older_version(self, last_msg_bytes):
        """
        Converts last message bytes of older version (version of converter) to current protocol version

        :param last_msg_bytes: first message bytes in older protocol version
        :return: converted message bytes
        """

    @abstractmethod
    def get_message_size_change_to_older_version(self):
        """
        Returns the difference in size between current protocol version and older version

        :return: message size difference
        """

    @abstractmethod
    def get_message_size_change_from_older_version(self):
        """
        Returns the difference in size between older version and current protocol version

        :return: message size difference
        """



from abc import ABCMeta
from asyncio import Task
from typing import Dict, Callable, Optional, Generic, TypeVar

from bxcommon.rpc.provider.abstract_provider import SubscriptionNotification

from bxcommon.constants import LOCALHOST, WS_DEFAULT_PORT
from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxcommon.rpc.ws.ws_client import WsClient
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType


logger = logging.get_logger(__name__)
memory_logger = logging.get_logger(LogRecordType.BxMemory, __name__)
msg_handling_logger = logging.get_logger(LogRecordType.MessageHandlingTroubleshooting, __name__)

T = TypeVar("T")


class AbstractFeedConnection(Generic[T], metaclass=ABCMeta):
    def __init__(
        self,
        node: T,
        feed_ip: str = LOCALHOST,
        feed_port: int = WS_DEFAULT_PORT,
        headers: Optional[Dict] = None
    ) -> None:
        self.node = node
        self.feed_ip = feed_ip
        self.feed_port = feed_port

        self.ws_client = WsClient(
            f"ws://{self.feed_ip}:{self.feed_port}/ws",
            headers=headers,
            retry_connection=True,
            retry_callback=self._subscribe_to_feeds
        )
        self.feeds_process: Dict[str, Callable[[SubscriptionNotification], None]] = {}
        self.subscription_tasks: Dict[str, Task] = {}

    async def subscribe_feeds(self) -> None:
        await self.ws_client.initialize()
        await self._subscribe_to_feeds(self.ws_client)

    async def _subscribe_to_feeds(self, ws_client: AbstractWsProvider) -> None:
        for feed_name, process in self.feeds_process.items():
            if feed_name in self.subscription_tasks:
                if self.subscription_tasks[feed_name].done():
                    logger.debug("FeedConnection resubscribed to feed {} from source {}", feed_name, self.feed_ip)
                    self.subscription_tasks[feed_name] = ws_client.subscribe_with_callback(process, feed_name)
            else:
                logger.debug("FeedConnection subscribed to feed {} from source {}", feed_name, self.feed_ip)
                self.subscription_tasks[feed_name] = ws_client.subscribe_with_callback(process, feed_name)

    async def revive(self) -> None:
        if self.ws_client.ws is None or not self.ws_client.running:
            logger.info("Attempting to revive websockets source feed...")
            await self.ws_client.connected_event.wait()

        if self.ws_client.ws is None or not self.ws_client.running:
            await self._subscribe_to_feeds(self.ws_client)

    async def stop(self) -> None:
        for _, task in self.subscription_tasks.items():
            task.cancel()
        await self.ws_client.close()



import struct

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxgateway.services.neutrality_service import NeutralityService
from bxutils import logging

logger = logging.get_logger(__name__)


class OntNeutralityService(NeutralityService):
    def _propagate_unencrypted_block_to_network(self, bx_block, connection, block_info):
        if block_info is None:
            raise ValueError("Block info is required to propagate unencrypted block")

        is_consensus_msg, = struct.unpack_from("?", bx_block[8:9])
        broadcast_type = BroadcastMessageType.CONSENSUS if is_consensus_msg else BroadcastMessageType.BLOCK

        broadcast_message = BroadcastMessage(block_info.block_hash, self._node.network_num,
                                             broadcast_type=broadcast_type, is_encrypted=False, blob=bx_block)

        conns = self._node.broadcast(
            broadcast_message,
            connection,
            connection_types=(ConnectionType.RELAY_BLOCK,)
        )
        handling_duration = self._node.track_block_from_node_handling_ended(block_info.block_hash)
        block_stats.add_block_event_by_block_hash(block_info.block_hash,
                                                  BlockStatEventType.ENC_BLOCK_SENT_FROM_GATEWAY_TO_NETWORK,
                                                  network_num=self._node.network_num,
                                                  broadcast_type=broadcast_type,
                                                  requested_by_peer=False,
                                                  peers=conns,
                                                  more_info="Peers: {}; Unencrypted; {}; Handled in {}".format(
                                                      stats_format.connections(conns),
                                                      self._format_block_info_stats(block_info),
                                                      stats_format.duration(handling_duration)))
        logger.info("Propagating block {} to the BDN.", block_info.block_hash)
        return broadcast_message



from dataclasses import dataclass

from bxcommon.models.config.abstract_config_model import AbstractConfigModel
from bxcommon.models.config.cron_config_model import CronConfigModel
from bxcommon.models.config.log_config_model import LogConfigModel
from bxcommon.models.config.stats_config_model import StatsConfigModel


@dataclass
class GatewayNodeConfigModel(AbstractConfigModel):
    # pyre-fixme[8]: Attribute has type `LogConfigModel`; used as `None`.
    log_config: LogConfigModel = None
    # pyre-fixme[8]: Attribute has type `StatsConfigModel`; used as `None`.
    stats_config: StatsConfigModel = None
    # pyre-fixme[8]: Attribute has type `CronConfigModel`; used as `None`.
    cron_config: CronConfigModel = None



from typing import List, Optional, Set

from bxcommon import constants
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.utils import ping_latency
from bxcommon.utils.ping_latency import NodeLatencyInfo
from bxutils import logging

logger = logging.get_logger(__name__)


def get_best_relays_by_ping_latency_one_per_country(
    relays: List[OutboundPeerModel], countries_count: int, current_relays: Optional[Set[OutboundPeerModel]] = None
) -> List[OutboundPeerModel]:
    """
    get best relay by pinging each relay and check its latency calculate with its inbound peers
    :param relays: list of relays
    :param countries_count: number of countries to return best relay for (one per country)
    :param current_relays: (for gateway use only) set of relays that gateway is currently connected to
    :return:
    """
    if len(relays) == 1:
        logger.debug("First (and only) recommended relay from api is: {}", relays[0])
        return relays

    relays_ping_latency = ping_latency.get_ping_latencies(relays)
    sorted_ping_latencies = sorted(relays_ping_latency, key=lambda relay_ts: relay_ts.latency)
    best_relays_by_latency = _get_best_relay_latencies_one_per_country(sorted_ping_latencies, relays, countries_count)
    logger.trace("Best relays by latency by country ({}): {}", countries_count, best_relays_by_latency)

    # if the absolute between fastest relay and recommended relay latency is less than NODE_LATENCY_THRESHOLD_MS,
    # select api's relay
    first_recommended_relay_latency = _get_node_latency_from_list(relays_ping_latency, relays[0].node_id)
    if (
        abs(best_relays_by_latency[0].latency - first_recommended_relay_latency)
        < constants.NODE_LATENCY_THRESHOLD_MS
    ):
        best_relay_node = relays[0]
    else:
        best_relay_node = best_relays_by_latency[0].node

    # for gateways only
    if current_relays is not None and len(current_relays) == 1 and best_relay_node not in current_relays:
        current_relay = next(iter(current_relays))
        if current_relay in relays:
            current_relay_latency = _get_node_latency_from_list(relays_ping_latency, current_relay.node_id)
            best_relay_node_latency = _get_node_latency_from_list(relays_ping_latency, best_relay_node.node_id)
            if current_relay_latency - best_relay_node_latency < constants.GATEWAY_SWAP_RELAYS_LATENCY_THRESHOLD_MS:
                best_relay_node = current_relay

    logger.info("Latency results for potential relays: [{}]",
                ", ".join([f"{relay_latency.node.ip} with {_format_latency(relay_latency.latency)}" for relay_latency in
                          sorted_ping_latencies]))
    logger.info("BDN recommended relay {} with {}. Node selected {}",
                relays[0].ip,
                "".join([_format_latency(relay.latency) for relay in relays_ping_latency if relay.node == relays[0]]),
                best_relay_node.ip)

    result_relays = [best_relay_node]

    if countries_count > 1:
        for extra_relay_latencies in best_relays_by_latency:
            if extra_relay_latencies.node.node_id != best_relay_node.node_id and \
                    extra_relay_latencies.node.get_country() != best_relay_node.get_country():
                result_relays.append(extra_relay_latencies.node)

    if len(result_relays) > 1:
        logger.info("Selected backup relays: {}",
                    ", ".join([f"{backup.ip} from {backup.get_country()}" for backup in result_relays[1:]]))

    return result_relays


def _get_best_relay_latencies_one_per_country(
        sorted_ping_latencies: List[NodeLatencyInfo], relays: List[OutboundPeerModel], countries_count: int
) -> List[NodeLatencyInfo]:
    fastest_latencies_by_country = {}
    best_latencies_by_country = {}
    ping_latency_threshold_by_country = {}

    for relay_latency in sorted_ping_latencies:
        relay = relay_latency.node
        relay_country = relay.get_country()

        if relay_country not in fastest_latencies_by_country:
            fastest_latencies_by_country[relay_country] = relay_latency
            best_latencies_by_country[relay_country] = relay_latency
            ping_latency_threshold_by_country[relay_country] = \
                constants.FASTEST_PING_LATENCY_THRESHOLD_PERCENT * relay_latency.latency
        elif relay_latency.latency - fastest_latencies_by_country[relay_country].latency < \
                ping_latency_threshold_by_country[relay_country] and \
                relays.index(relay_latency.node) < relays.index(best_latencies_by_country[relay_country].node):
            best_latencies_by_country[relay_country] = relay_latency

    sorted_latencies_by_country = list(
        map(
                lambda dic_item: dic_item[1],
                sorted(best_latencies_by_country.items(), key=lambda dic_item: dic_item[1].latency)
            )
    )

    if countries_count > len(sorted_latencies_by_country):
        return sorted_latencies_by_country
    else:
        return sorted_latencies_by_country[:countries_count]


def _format_latency(latency_ms: float) -> str:
    if latency_ms < constants.PING_TIMEOUT_S * 2000:
        return f"latency {latency_ms} ms"

    return "latency timeout"


def _get_node_latency_from_list(node_latencies: List[NodeLatencyInfo], node_id: Optional[str]) -> float:
    for node_latency in node_latencies:
        if node_latency.node.node_id == node_id:
            return node_latency.latency
    return constants.PING_INTERVAL_S * 1000



from typing import Optional, Tuple, Dict, Any

from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


class EthWsSubscriber(AbstractWsProvider):
    """
    Subscriber to Ethereum websockets RPC interface.

    Requires Ethereum startup arguments:
    --ws --wsapi eth --wsport 8546

    See https://geth.ethereum.org/docs/rpc/server for more info.
    (--ws-addr 127.0.0.1 maybe a good idea too)

    Requires Gateway startup arguments:
    --eth-ws-uri ws://127.0.0.1:8546
    """

    def __init__(
        self,
        ws_uri: str,
    ) -> None:
        super().__init__(ws_uri, True)

    async def reconnect(self) -> None:
        logger.warning(log_messages.ETH_WS_SUBSCRIBER_CONNECTION_BROKEN)
        try:
            await super().reconnect()
        except ConnectionRefusedError:
            self.running = False

        if self.running:
            logger.info("Reconnected to Ethereum websocket feed")
        else:
            logger.warning(log_messages.ETH_RPC_COULD_NOT_RECONNECT)

    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        response = await self.call_rpc(
            "eth_subscribe",
            [channel]
        )
        subscription_id = response.result
        assert isinstance(subscription_id, str)
        self.subscription_manager.register_subscription(subscription_id)
        return subscription_id

    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        response = await self.call_rpc(
            "eth_unsubscribe",
            [subscription_id]
        )
        if response.result is not None:
            self.subscription_manager.unregister_subscription(subscription_id)
            return True, None
        else:
            return False, response.error



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.protocol_version import PROTOCOL_VERSION
from bxcommon.utils.message_buffer_builder import PayloadElement, PayloadBlock


class KeepAliveMessage(AbstractBloxrouteMessage):
    """
    BloXroute Version message that contains a message nonce to be used in the response.

    nonce: long, to be provided and managed by the connection
    """
    KEEP_ALIVE_MESSAGE_BLOCK = PayloadBlock(
        AbstractBloxrouteMessage.HEADER_LENGTH, "ResponseMessage", PROTOCOL_VERSION,
        PayloadElement(name="nonce", structure="<Q", decode=lambda x: x or None),
    )
    KEEP_ALIVE_MESSAGE_LENGTH = KEEP_ALIVE_MESSAGE_BLOCK.size + constants.CONTROL_FLAGS_LEN

    def __init__(
        self,
        msg_type: bytes,
        nonce: Optional[int] = None,
        buf: Optional[bytearray] = None,
        payload_length: Optional[int] = None
    ) -> None:
        if payload_length is None:
            payload_length = self.KEEP_ALIVE_MESSAGE_LENGTH
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + payload_length)

        buf = self.KEEP_ALIVE_MESSAGE_BLOCK.build(buf, nonce=nonce)

        self.buf = buf
        self._nonce: Optional[int] = None
        self._network_num: Optional[int] = None
        self._memoryview = memoryview(buf)
        super(KeepAliveMessage, self).__init__(msg_type, payload_length, buf)

    def __unpack(self) -> None:
        contents = self.KEEP_ALIVE_MESSAGE_BLOCK.read(self._memoryview)
        self._nonce = contents.get("nonce")

    def nonce(self) -> int:
        if self._nonce is None:
            self.__unpack()
        nonce = self._nonce
        if nonce is None:
            return 0
        return nonce



import time
from asyncio import BufferedProtocol
from typing import Optional, TYPE_CHECKING

from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxutils import logging
from bxutils.logging import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
network_troubleshooting_logger = logging.get_logger(LogRecordType.NetworkTroubleshooting, __name__)


# pyre-fixme[11]: Annotation `BufferedProtocol` is not defined as a type.
class SocketConnectionProtocol(AbstractSocketConnectionProtocol, BufferedProtocol):
    def __init__(
        self,
        node: "AbstractNode",
        endpoint: Optional[IpEndpoint] = None,
        is_ssl: bool = True,
    ):
        AbstractSocketConnectionProtocol.__init__(self, node, endpoint, is_ssl)
        self._buffer_request_time: Optional[float] = None
        self._buffer_update_time: Optional[float] = None

    # pylint: disable=arguments-differ
    def get_buffer(self, _sizehint: int):
        self._buffer_request_time = time.time()
        logger.trace("[{}] - get_buffer {}.", self, _sizehint)
        return self._receive_buf

    def buffer_updated(self, nbytes: int) -> None:
        if self.is_receivable():
            self._buffer_update_time = time.time()
            logger.trace("[{}] - buffer_updated {}.", self, nbytes)
            self._node.on_bytes_received(self.file_no, self._receive_buf[:nbytes])

    def get_last_read_duration_ms(self) -> float:
        if self._buffer_request_time and self._buffer_update_time:
            end_time = self._buffer_update_time
            start_time = self._buffer_request_time
            assert end_time is not None
            assert start_time is not None
            return (end_time - start_time) * 1000

        return 0

    def get_time_since_read_end_ms(self, end_time: float) -> float:
        if self._buffer_update_time:
            start_time = self._buffer_request_time
            assert start_time is not None
            return (end_time - start_time) * 1000

        return 0



from typing import TYPE_CHECKING

from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.utils import memory_utils
from bxcommon.utils.stats import stats_format

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

TOTAL_MEM_USAGE = "total_mem_usage"
TOTAL_CACHED_TX = "total_cached_transactions"
TOTAL_CACHED_TX_SIZE = "total_cached_transactions_size"


class GatewayMemoryRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):

    help = {
        "params": "",
        "description": "return gateway node memory information"
    }

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        tx_service = self.node.get_tx_service()
        cache_state = tx_service.get_cache_state_json()
        return self.ok({
            TOTAL_MEM_USAGE: stats_format.byte_count(memory_utils.get_app_memory_usage()),
            TOTAL_CACHED_TX: cache_state["tx_hash_to_contents_len"],
            TOTAL_CACHED_TX_SIZE: stats_format.byte_count(cache_state["total_tx_contents_size"])
        })




import os
import sys

from argparse import Namespace
from dataclasses import dataclass
from typing import Optional, Set, Dict

from bxcommon import constants
from bxcommon.common_opts import CommonOpts
from bxcommon.rpc import rpc_constants
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils import ip_resolver, node_cache
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxcommon.utils.convert import hex_to_bytes
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxgateway import argument_parsers
from bxgateway import gateway_constants
from bxgateway import log_messages
from bxgateway.utils.eth.eccx import ECCx
from bxutils import logging


logger = logging.get_logger(__name__)


@dataclass
class GatewayOpts(CommonOpts):
    blockchain_port: int
    blockchain_protocol: Optional[str]
    blockchain_network: Optional[str]
    blockchain_networks: Dict[int, BlockchainNetworkModel]
    blockchain_block_recovery_timeout_s: int
    blockchain_block_hold_timeout_s: int
    blockchain_block_interval: int
    blockchain_ignore_block_interval_count: int
    blockchain_ip: str
    peer_gateways: Set[OutboundPeerModel]
    peer_transaction_relays: Set[OutboundPeerModel]
    remote_blockchain_peer: OutboundPeerModel
    min_peer_gateways: int
    remote_blockchain_ip: str
    remote_blockchain_port: int
    connect_to_remote_blockchain: bool
    encrypt_blocks: bool
    peer_relays: Set[OutboundPeerModel]
    test_mode: str
    blockchain_version: int
    blockchain_nonce: int
    blockchain_net_magic: int
    blockchain_services: int
    enable_node_cache: bool
    node_public_key: str
    enode: str
    private_key: str
    network_id: int
    genesis_hash: str
    chain_difficulty: str
    no_discovery: bool
    remote_public_key: str
    blockchain_peers: Set[BlockchainPeerInfo]
    blockchain_peers_file: str
    compact_block: bool
    compact_block_min_tx_count: int
    dump_short_id_mapping_compression: bool
    dump_short_id_mapping_compression_path: str
    tune_send_buffer_size: bool
    max_block_interval_s: int
    cookie_file_path: str
    blockchain_message_ttl: int
    remote_blockchain_message_ttl: int
    stay_alive_duration: int
    initial_liveliness_check: int
    config_update_interval: int
    require_blockchain_connection: bool
    default_tx_flag: TransactionFlag
    should_update_source_version: bool
    account_model: Optional[BdnAccountModelBase]
    process_node_txs_in_extension: bool
    enable_eth_extensions: bool     # TODO remove
    request_recovery: bool
    enable_block_compression: bool
    filter_txs_factor: float
    min_peer_relays_count: int
    should_restart_on_high_memory: bool

    # IPC
    ipc: bool
    ipc_file: str

    # Ontology specific
    http_info_port: int
    consensus_port: int
    relay: bool
    is_consensus: bool

    # transaction feed
    ws: bool
    ws_host: str
    ws_port: int
    eth_ws_uri: Optional[str]
    request_remote_transaction_streaming: bool
    stream_to_peer_gateway: Optional[OutboundPeerModel]

    # ENV
    is_docker: bool

    @classmethod
    def opts_defaults(cls, opts) -> Namespace:
        opts = super().opts_defaults(opts)

        if "blockchain_networks" not in opts:
            # node_cache dependencies should be untangled
            #  parameter to call `node_cache.read` but got `Namespace`
            cache_file_info = node_cache.read(opts)
            if cache_file_info is not None:
                opts.blockchain_networks = cache_file_info.blockchain_networks

        opts.outbound_peers = set(opts.peer_gateways).union(opts.peer_relays)

        if opts.connect_to_remote_blockchain and opts.remote_blockchain_ip and opts.remote_blockchain_port:
            opts.remote_blockchain_peer = OutboundPeerModel(opts.remote_blockchain_ip, opts.remote_blockchain_port)
        else:
            opts.remote_blockchain_peer = None

        opts.account_model = None

        opts.is_docker = os.path.exists("/.dockerenv")

        # Request streaming from BDN if ws server is turned on
        opts.request_remote_transaction_streaming = opts.ws

        # set by node runner
        opts.blockchain_block_interval = 0
        opts.blockchain_ignore_block_interval_count = 0
        opts.blockchain_block_recovery_timeout_s = 0
        opts.blockchain_block_hold_timeout_s = 0
        opts.enable_network_content_logs = False
        opts.should_update_source_version = False

        # set after initialization
        opts.peer_transaction_relays = []

        if opts.blockchain_protocol:
            opts.blockchain_protocol = opts.blockchain_protocol.lower()
        else:
            opts.blockchain_protocol = None

        if not opts.blockchain_peers:
            opts.blockchain_peers = set()

        if opts.blockchain_peers_file:
            cls.read_blockchain_peers_from_file(opts)

        if not opts.cookie_file_path:
            opts.cookie_file_path = gateway_constants.COOKIE_FILE_PATH_TEMPLATE.format(
                "{}_{}".format(get_sdn_hostname(opts.sdn_url), opts.external_port)
            )

        opts.min_peer_relays_count = 1

        if opts.stream_to_peer_gateway is not None:
            opts.peer_gateways.append(opts.stream_to_peer_gateway)
        return opts

    @classmethod
    def read_blockchain_peers_from_file(cls, opts) -> None:
        try:
            with open(opts.blockchain_peers_file, "r", encoding=constants.DEFAULT_TEXT_ENCODING) as peers_file:
                blockchain_peers_list = peers_file.readlines()
                blockchain_peers_list = [peer.strip() for peer in blockchain_peers_list]
                blockchain_peers_list = [peer for peer in blockchain_peers_list if peer]
                for peer in blockchain_peers_list:
                    opts.blockchain_peers.add(argument_parsers.parse_peer(opts.blockchain_protocol, peer))
        except FileNotFoundError:
            logger.fatal(log_messages.BLOCKCHAIN_PEERS_FILE_NOT_FOUND, opts.blockchain_peers_file, exc_info=False)
            sys.exit(1)

    def __post_init__(self):
        if self.filter_txs_factor < 0:
            logger.fatal("--filter_txs_factor cannot be below 0.")
            sys.exit(1)

    def validate_eth_opts(self) -> None:
        if not self.blockchain_ip and not self.blockchain_peers:
            logger.fatal(log_messages.ETH_MISSING_BLOCKCHAIN_IP_AND_BLOCKCHAIN_PEERS, exc_info=False)
            sys.exit(1)

        if self.blockchain_ip:
            if self.node_public_key is None:
                logger.fatal(log_messages.ETH_MISSING_NODE_PUBLIC_KEY, exc_info=False)
                sys.exit(1)
            validate_pub_key(self.node_public_key)

        if self.blockchain_peers:
            for blockchain_peer in self.blockchain_peers:
                if blockchain_peer.ip is None:
                    logger.fatal(log_messages.ETH_MISSING_BLOCKCHAIN_IP_FROM_BLOCKCHAIN_PEERS, exc_info=False)
                    sys.exit(1)
                if blockchain_peer.node_public_key is None:
                    logger.fatal(log_messages.ETH_MISSING_NODE_PUBLIC_KEY_FROM_BLOCKCHAIN_PEERS, exc_info=False)
                    sys.exit(1)
                validate_pub_key(blockchain_peer.node_public_key)

        if self.remote_blockchain_peer is not None:
            if self.remote_public_key is None:
                logger.fatal(log_messages.ETH_MISSING_REMOTE_NODE_PUBLIC_KEY, exc_info=False)
                sys.exit(1)
            validate_pub_key(self.remote_public_key)

    def set_account_options(self, account_model: BdnAccountModelBase) -> None:
        super().set_account_options(account_model)
        self.account_model = account_model

        blockchain_protocol = account_model.blockchain_protocol
        blockchain_network = account_model.blockchain_network
        if self.blockchain_protocol is None and blockchain_protocol is not None:
            self.blockchain_protocol = blockchain_protocol.lower()
        if self.blockchain_network is None and blockchain_network is not None:
            self.blockchain_network = blockchain_network

    def validate_network_opts(self) -> None:
        if self.blockchain_network is None:
            self.blockchain_network = "mainnet"

        blockchain_protocol = self.blockchain_protocol

        if blockchain_protocol is None:
            logger.fatal(log_messages.MISSING_BLOCKCHAIN_PROTOCOL)
            sys.exit(1)

        if blockchain_protocol == BlockchainProtocol.ETHEREUM:
            self.validate_eth_opts()

        if not self.blockchain_ip and not self.blockchain_peers:
            logger.fatal(log_messages.MISSING_BLOCKCHAIN_IP_AND_BLOCKCHAIN_PEERS, exc_info=False)
            sys.exit(1)

        if self.blockchain_ip:
            self.blockchain_ip = validate_blockchain_ip(self.blockchain_ip, self.is_docker)
        if self.blockchain_peers:
            for blockchain_peer in self.blockchain_peers:
                blockchain_peer.ip = validate_blockchain_ip(blockchain_peer.ip, self.is_docker)
        if self.rpc_host == rpc_constants.DEFAULT_RPC_HOST and self.is_docker:
            docker_host = get_docker_host()
            if docker_host:
                self.rpc_host = docker_host
            else:
                self.rpc_host = rpc_constants.DEFAULT_RPC_HOST


def get_sdn_hostname(sdn_url: str) -> str:
    new_sdn_url = sdn_url
    if "://" in sdn_url:
        new_sdn_url = sdn_url.split("://")[1]

    return new_sdn_url


def get_docker_host() -> Optional[str]:
    with open('/etc/hosts', "r") as f:
        host_list = f.readlines()
        for host in host_list:
            if "172" in host:
                end_idx = 0
                for char in host:
                    if not char.isnumeric() and char != '.':
                        return host[:end_idx]
                    end_idx += 1


def validate_pub_key(key) -> None:
    if key.startswith("0x"):
        key = key[2:]
    if len(key) != 2 * eth_common_constants.PUBLIC_KEY_LEN:
        logger.fatal(log_messages.INVALID_PUBLIC_KEY_LENGTH,
                     len(key), exc_info=False)
        sys.exit(1)
    eccx_obj = ECCx()
    if not eccx_obj.is_valid_key(hex_to_bytes(key)):
        logger.fatal(log_messages.INVALID_PUBLIC_KEY, exc_info=False)
        sys.exit(1)


def validate_blockchain_ip(blockchain_ip, is_docker=False) -> str:
    if blockchain_ip is None:
        logger.fatal(log_messages.MISSING_BLOCKCHAIN_IP_FROM_BLOCKCHAIN_PEERS, exc_info=False)
        sys.exit(1)

    if blockchain_ip == gateway_constants.LOCALHOST and is_docker:
        logger.warning(log_messages.INVALID_BLOCKCHAIN_IP, exc_info=False)
    try:
        return ip_resolver.blocking_resolve_ip(blockchain_ip)
    except EnvironmentError:
        logger.fatal(log_messages.BLOCKCHAIN_IP_RESOLVE_ERROR, blockchain_ip)
        sys.exit(1)




from cryptography import utils
from cryptography.x509.extensions import UnrecognizedExtension

from bxutils.ssl.extensions.extensions_object_ids import ExtensionsObjectIds


class NodeIdExtension(UnrecognizedExtension):

    def __init__(self, node_id: str) -> None:
        super(NodeIdExtension, self).__init__(ExtensionsObjectIds.NODE_ID, node_id.encode("utf-8"))
        self._node_id = node_id

    node_id = utils.read_only_property("_node_id")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <node_id: {self._node_id}>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, NodeIdExtension):
            return NotImplemented

        return self._node_id == other._node_id

    def __ne__(self, other) -> bool:
        return not self == other

    def __hash__(self):
        return hash(self._node_id)
